import os

import cv2
import random
import numpy as np


def tens2image(im):
    if im.size()[0] == 1:
        tmp = np.squeeze(im.numpy(), axis=0)
    else:
        tmp = im.numpy()
    if tmp.ndim == 2:
        return tmp
    else:
        return tmp.transpose((1, 2, 0))


def crop2fullmask(crop_mask, bbox, im=None, im_size=None, zero_pad=False, relax=0, mask_relax=True,
                  interpolation=cv2.INTER_CUBIC, scikit=False):
    if scikit:
        from skimage.transform import resize as sk_resize
    assert(not(im is None and im_size is None)), 'You have to provide an image or the image size'
    if im is None:
        im_si = im_size
    else:
        im_si = im.shape
    # Borers of image
    bounds = (0, 0, im_si[1] - 1, im_si[0] - 1)

    # Valid bounding box locations as (x_min, y_min, x_max, y_max)
    bbox_valid = (max(bbox[0], bounds[0]),
                  max(bbox[1], bounds[1]),
                  min(bbox[2], bounds[2]),
                  min(bbox[3], bounds[3]))

    # Bounding box of initial mask
    bbox_init = (bbox[0] + relax,
                 bbox[1] + relax,
                 bbox[2] - relax,
                 bbox[3] - relax)

    if zero_pad:
        # Offsets for x and y
        offsets = (-bbox[0], -bbox[1])
    else:
        assert((bbox == bbox_valid).all())
        offsets = (-bbox_valid[0], -bbox_valid[1])

    # Simple per element addition in the tuple
    inds = tuple(map(sum, zip(bbox_valid, offsets + offsets)))

    if scikit:
        crop_mask = sk_resize(crop_mask, (bbox[3] - bbox[1] + 1, bbox[2] - bbox[0] + 1), order=0, mode='constant').astype(crop_mask.dtype)
    else:
        crop_mask = cv2.resize(crop_mask, (bbox[2] - bbox[0] + 1, bbox[3] - bbox[1] + 1), interpolation=interpolation)
    result_ = np.zeros(im_si)
    result_[bbox_valid[1]:bbox_valid[3] + 1, bbox_valid[0]:bbox_valid[2] + 1] = \
        crop_mask[inds[1]:inds[3] + 1, inds[0]:inds[2] + 1]

    result = np.zeros(im_si)
    if mask_relax:
        result[bbox_init[1]:bbox_init[3]+1, bbox_init[0]:bbox_init[2]+1] = \
            result_[bbox_init[1]:bbox_init[3]+1, bbox_init[0]:bbox_init[2]+1]
    else:
        result = result_

    return result


def overlay_mask(im, ma, colors=None, alpha=0.5):
    assert np.max(im) <= 1.0
    if colors is None:
        colors = np.load(os.path.join(os.path.dirname(__file__), 'pascal_map.npy'))/255.
    else:
        colors = np.append([[0.,0.,0.]], colors, axis=0);

    if ma.ndim == 3:
        assert len(colors) >= ma.shape[0], 'Not enough colors'
    ma = ma.astype(np.bool)
    im = im.astype(np.float32)

    if ma.ndim == 2:
        fg = im * alpha+np.ones(im.shape) * (1 - alpha) * colors[1, :3]   # np.array([0,0,255])/255.0
    else:
        fg = []
        for n in range(ma.ndim):
            fg.append(im * alpha + np.ones(im.shape) * (1 - alpha) * colors[1+n, :3])
    # Whiten background
    bg = im.copy()
    if ma.ndim == 2:
        bg[ma == 0] = im[ma == 0]
        bg[ma == 1] = fg[ma == 1]
        total_ma = ma
    else:
        total_ma = np.zeros([ma.shape[1], ma.shape[2]])
        for n in range(ma.shape[0]):
            tmp_ma = ma[n, :, :]
            total_ma = np.logical_or(tmp_ma, total_ma)
            tmp_fg = fg[n]
            bg[tmp_ma == 1] = tmp_fg[tmp_ma == 1]
        bg[total_ma == 0] = im[total_ma == 0]

    # [-2:] is s trick to be compatible both with opencv 2 and 3
    contours = cv2.findContours(total_ma.copy().astype(np.uint8), cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)[-2:]
    cv2.drawContours(bg, contours[0], -1, (0.0, 0.0, 0.0), 1)

    return bg


def overlay_masks(im, masks, alpha=0.5):
    colors = np.load(os.path.join(os.path.dirname(__file__), 'pascal_map.npy'))/255.
    
    if isinstance(masks, np.ndarray):
        masks = [masks]

    assert len(colors) >= len(masks), 'Not enough colors'

    ov = im.copy()
    im = im.astype(np.float32)
    total_ma = np.zeros([im.shape[0], im.shape[1]])
    i = 1
    for ma in masks:
        ma = ma.astype(np.bool)
        fg = im * alpha+np.ones(im.shape) * (1 - alpha) * colors[i, :3]   # np.array([0,0,255])/255.0
        i = i + 1
        ov[ma == 1] = fg[ma == 1]
        total_ma += ma

        # [-2:] is s trick to be compatible both with opencv 2 and 3
        contours = cv2.findContours(ma.copy().astype(np.uint8), cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)[-2:]
        cv2.drawContours(ov, contours[0], -1, (0.0, 0.0, 0.0), 1)
    ov[total_ma == 0] = im[total_ma == 0]

    return ov


def extreme_points(mask, pert):
    def find_point(id_x, id_y, ids):
        sel_id = ids[0][random.randint(0, len(ids[0]) - 1)]
        return [id_x[sel_id], id_y[sel_id]]

    # List of coordinates of the mask
    inds_y, inds_x = np.where(mask > 0.5)

    # Find extreme points
    return np.array([find_point(inds_x, inds_y, np.where(inds_x <= np.min(inds_x)+pert)), # left
                     find_point(inds_x, inds_y, np.where(inds_x >= np.max(inds_x)-pert)), # right
                     find_point(inds_x, inds_y, np.where(inds_y <= np.min(inds_y)+pert)), # top
                     find_point(inds_x, inds_y, np.where(inds_y >= np.max(inds_y)-pert)) # bottom
                     ])


def get_bbox(mask, points=None, pad=0, zero_pad=False):
    if points is not None:
        inds = np.flip(points.transpose(), axis=0)
    else:
        inds = np.where(mask > 0)

    if inds[0].shape[0] == 0:
        return None

    if zero_pad:
        x_min_bound = -np.inf
        y_min_bound = -np.inf
        x_max_bound = np.inf
        y_max_bound = np.inf
    else:
        x_min_bound = 0
        y_min_bound = 0
        x_max_bound = mask.shape[1] - 1
        y_max_bound = mask.shape[0] - 1

    x_min = max(inds[1].min() - pad, x_min_bound)
    y_min = max(inds[0].min() - pad, y_min_bound)
    x_max = min(inds[1].max() + pad, x_max_bound)
    y_max = min(inds[0].max() + pad, y_max_bound)

    return x_min, y_min, x_max, y_max


def crop_from_bbox(img, bbox, zero_pad=False):
    # Borders of image
    bounds = (0, 0, img.shape[1] - 1, img.shape[0] - 1)

    # Valid bounding box locations as (x_min, y_min, x_max, y_max)
    bbox_valid = (max(bbox[0], bounds[0]),
                  max(bbox[1], bounds[1]),
                  min(bbox[2], bounds[2]),
                  min(bbox[3], bounds[3]))

    if zero_pad:
        # Initialize crop size (first 2 dimensions)
        crop = np.zeros((bbox[3] - bbox[1] + 1, bbox[2] - bbox[0] + 1), dtype=img.dtype)

        # Offsets for x and y
        offsets = (-bbox[0], -bbox[1])

    else:
        assert(bbox == bbox_valid)
        crop = np.zeros((bbox_valid[3] - bbox_valid[1] + 1, bbox_valid[2] - bbox_valid[0] + 1), dtype=img.dtype)
        offsets = (-bbox_valid[0], -bbox_valid[1])

    # Simple per element addition in the tuple
    inds = tuple(map(sum, zip(bbox_valid, offsets + offsets)))

    img = np.squeeze(img)
    if img.ndim == 2:
        crop[inds[1]:inds[3] + 1, inds[0]:inds[2] + 1] = \
            img[bbox_valid[1]:bbox_valid[3] + 1, bbox_valid[0]:bbox_valid[2] + 1]
    else:
        crop = np.tile(crop[:, :, np.newaxis], [1, 1, 3])  # Add 3 RGB Channels
        crop[inds[1]:inds[3] + 1, inds[0]:inds[2] + 1, :] = \
            img[bbox_valid[1]:bbox_valid[3] + 1, bbox_valid[0]:bbox_valid[2] + 1, :]

    return crop


def fixed_resize(sample, resolution, flagval=None):

    if flagval is None:
        if ((sample == 0) | (sample == 1)).all():
            flagval = cv2.INTER_NEAREST
        else:
            flagval = cv2.INTER_CUBIC

    if isinstance(resolution, int):
        tmp = [resolution, resolution]
        tmp[np.argmax(sample.shape[:2])] = int(round(float(resolution)/np.min(sample.shape[:2])*np.max(sample.shape[:2])))
        resolution = tuple(tmp)

    if sample.ndim == 2 or (sample.ndim == 3 and sample.shape[2] == 3):
        sample = cv2.resize(sample, resolution[::-1], interpolation=flagval)
    else:
        tmp = sample
        sample = np.zeros(np.append(resolution, tmp.shape[2]), dtype=np.float32)
        for ii in range(sample.shape[2]):
            sample[:, :, ii] = cv2.resize(tmp[:, :, ii], resolution[::-1], interpolation=flagval)
    return sample


def crop_from_mask(img, mask, relax=0, zero_pad=False):
    if mask.shape[:2] != img.shape[:2]:
        mask = cv2.resize(mask, dsize=tuple(reversed(img.shape[:2])), interpolation=cv2.INTER_NEAREST)

    assert(mask.shape[:2] == img.shape[:2])

    bbox = get_bbox(mask, pad=relax, zero_pad=zero_pad)

    if bbox is None:
        return None

    crop = crop_from_bbox(img, bbox, zero_pad)

    return crop


def make_gaussian(size, sigma=10, center=None, d_type=np.float64):
    """ Make a square gaussian kernel.
    size: is the dimensions of the output gaussian
    sigma: is full-width-half-maximum, which
    can be thought of as an effective radius.
    """

    x = np.arange(0, size[1], 1, float)
    y = np.arange(0, size[0], 1, float)
    y = y[:, np.newaxis]

    if center is None:
        x0 = y0 = size[0] // 2
    else:
        x0 = center[0]
        y0 = center[1]

    return np.exp(-4 * np.log(2) * ((x - x0) ** 2 + (y - y0) ** 2) / sigma ** 2).astype(d_type)


def make_gt(img, labels, sigma=10, one_mask_per_point=False):
    """ Make the ground-truth for  landmark.
    img: the original color image
    labels: label with the Gaussian center(s) [[x0, y0],[x1, y1],...]
    sigma: sigma of the Gaussian.
    one_mask_per_point: masks for each point in different channels?
    """
    h, w = img.shape[:2]
    if labels is None:
        gt = make_gaussian((h, w), center=(h//2, w//2), sigma=sigma)
    else:
        labels = np.array(labels)
        if labels.ndim == 1:
            labels = labels[np.newaxis]
        if one_mask_per_point:
            gt = np.zeros(shape=(h, w, labels.shape[0]))
            for ii in range(labels.shape[0]):
                gt[:, :, ii] = make_gaussian((h, w), center=labels[ii, :], sigma=sigma)
        else:
            gt = np.zeros(shape=(h, w), dtype=np.float64)
            for ii in range(labels.shape[0]):
                gt = np.maximum(gt, make_gaussian((h, w), center=labels[ii, :], sigma=sigma))

    gt = gt.astype(dtype=img.dtype)

    return gt


def cstm_normalize(im, max_value):
    """
    Normalize image to range 0 - max_value
    """
    imn = max_value*(im - im.min()) / max((im.max() - im.min()), 1e-8)
    return imn


def generate_param_report(logfile, param):
    log_file = open(logfile, 'w')
    for key, val in param.items():
        log_file.write(key+':'+str(val)+'\n')
    log_file.close()



# -*- coding: utf-8 -*-
"""
Created on Wed Nov 13 13:12:11 2019

@author: Koerber
"""


import imgaug.augmenters as iaa
import numpy as np
import random
import math
import cv2
from dl.method.mode import dlMode

class ImageAugment():
    def __init__(self, parent):
        # this is the model input size
        self.parent = parent
        self.modelinputwidth = 224  
        self.modelinputheight = 224
        
        self.enableAll = True
        self.flip_horz = True
        self.flip_vert = True
        self.prob_gaussian = 0.4 
        self.prob_piecewise = 0.3
        self.prob_dropout = 0.3
        
        self.enableAffine = True
        self.prob_affine = 0.5
        self.scale_min = 0.8
        self.scale_max = 1.2
        self.translate_min = -0.2
        self.translate_max = 0.2
        self.rotate_min = -45
        self.rotate_max = 45
        self.shear_min = -16
        self.shear_max = 16
        
        self.seg_ignoreBackground = True
        self.seg_ignorecycles = 5
        self.seg_minRequiredPixels = 1
        
        self.od_ignoreBackground = True
        self.od_ignorecycles = 5
        self.od_minRequiredPixels = 1
        
        self.augmentation_sequence = None
        
        self.currentTile = 0


    @property
    def outputwidth(self):
        # it is called outputwidth because it is the output of the augmentation and the input to the model
        try:
            if self.parent.InputSize[1] is not None:
                return self.parent.InputSize[1]
        except:
            pass
        return self.modelinputwidth

    @outputwidth.setter
    def outputwidth(self, val):
        self.modelinputwidth = val


    @property
    def outputheight(self):
        # it is called outputwidth because it is the output of the augmentation and the input to the model
        try:
            if self.parent.InputSize[0] is not None:
                return self.parent.InputSize[0]
        except:
            pass
        return self.modelinputheight

    @outputheight.setter
    def outputheight(self, val):
        self.modelinputheight = val
        
    @property
    def ignoreBackground(self):
        if self.parent.WorkingMode == dlMode.Segmentation:
            return self.seg_ignoreBackground
        elif self.parent.WorkingMode == dlMode.Object_Counting:
            return self.od_ignoreBackground
        else:
            return False
        
    @property
    def ignorecycles(self):
        if self.parent.WorkingMode == dlMode.Segmentation:
            return self.seg_ignorecycles
        elif self.parent.WorkingMode == dlMode.Object_Counting:
            return self.od_ignorecycles
        else:
            return 0
    
    @property
    def minRequiredPixels(self):
        if self.parent.WorkingMode == dlMode.Segmentation:
            return self.seg_minRequiredPixels
        elif self.parent.WorkingMode == dlMode.Object_Counting:
            return self.od_minRequiredPixels
        else:
            return 1
        
    def setoutputwidth(self, w):
        self.outputwidth = w - w % self.parent.split_factor
        
    def setoutputheight(self, h):
        self.outputheight = h - h % self.parent.split_factor
        
    def randomcrop_or_pad(self, images, labels):
        # input a list of images
        # output a single array with images (num, outputwidth, outputheight)
        imgs = []
        labls = []
        for image,label in zip(images, labels):
            assert (label.shape[0:1] == image.shape[0:1])
            if image.shape[1] < self.outputwidth or image.shape[0] < self.outputheight:
                w = max(0,self.outputwidth - image.shape[1])
                h = max(0,self.outputheight - image.shape[0])
                # after padding size is increased by h or w in that direction
                pad = ((h, h), (w, w), (0, 0))
                image = np.pad(image, pad, 'reflect')
                label = np.pad(label, pad, 'reflect')

            if self.ignoreBackground:
                counter = 0
                while counter < self.ignorecycles:
                    x = random.randint(0, image.shape[1] - self.outputwidth)
                    y = random.randint(0, image.shape[0] - self.outputheight)
                    if np.count_nonzero(label[y:y+self.outputheight, x:x+self.outputwidth,...]) >= self.minRequiredPixels:
                        break
                    counter += 1
            else:
                x = random.randint(0, image.shape[1] - self.outputwidth)
                y = random.randint(0, image.shape[0] - self.outputheight)
                
            image = image[y:y+self.outputheight, x:x+self.outputwidth,...]
            label = label[y:y+self.outputheight, x:x+self.outputwidth,...]
 
            imgs.append(image)
            labls.append(label)
            
        return imgs, labls
    
    def getTilesPerImage(self, width, height):
        if self.parent.Mode.type == dlMode.Segmentation or self.parent.Mode.type == dlMode.Object_Counting:
            return  math.ceil(width/self.outputwidth) * math.ceil(height/self.outputheight)
        elif self.parent.Mode.type == dlMode.Classification:
            return 1
            
    
    def systematiccrop_or_pad(self, images, labels):     
        # images are cropped from left-to-right and top-to-bot      
        # self.currentTile keeps track of the position in the parent image 
        # it requires sorted images as input to return correct tiles
        imgs = []
        labls = []
        for image,label in zip(images, labels):           
            assert (label.shape[0:2] == image.shape[0:2])
            if image.shape[1] < self.outputwidth or image.shape[0] < self.outputheight:
                w = max(0,self.outputwidth - image.shape[1])
                h = max(0,self.outputheight - image.shape[0])
                # after padding size is increased by h or w in that direction
                pad = ((h, 0), (w, 0), (0, 0))
                image = np.pad(image, pad, 'reflect')
                label = np.pad(label, pad, 'reflect')

            
            num_in_width = math.ceil(image.shape[1] / self.outputwidth)
            num_in_height = math.ceil(image.shape[0] / self.outputheight)
            x = (self.currentTile % num_in_width) * self.outputwidth
            y = (self.currentTile // num_in_width) * self.outputheight
            
            if self.currentTile+1 >= self.getTilesPerImage(image.shape[1],image.shape[0]):
                self.currentTile = 0
            else:
                self.currentTile += 1
                
            x_start = min(x, image.shape[1] - self.outputwidth)
            y_start = min(y, image.shape[0] - self.outputheight)
            x_end = min(x+self.outputwidth, image.shape[1])
            y_end = min(y+self.outputheight, image.shape[0])
            image = image[y_start:y_end, x_start:x_end,...]
            label = label[y_start:y_end, x_start:x_end,...]
            
            
            imgs.append(image)
            labls.append(label)

        return imgs, labls
        
    def fitFullSizeImage2OutputSize(self, images):
        # add random crop?
        dims = images[0].shape
        ret = [cv2.resize(x, (self.outputwidth, self.outputheight), interpolation = cv2.INTER_CUBIC) for x in images]
        # in case of monochromatic
        if dims > ret[0].shape:
            ret = [x[...,np.newaxis] for x in ret]
        return ret

    def checkModelSizeCompatibility(self, image, label  = None):       
        targetsize = self.parent.InputSize
        currsize = image.shape[1:3]
        idims = len(image.shape)
        
        if targetsize[0] is None and targetsize[1] is None:
            targetsize = (self.outputheight, self.outputwidth)
        
        
        if (targetsize[0] is not None or targetsize[1] is not None) and (targetsize != currsize):
            image = np.stack([cv2.resize(x, (targetsize[1], targetsize[0]), interpolation = cv2.INTER_CUBIC) for x in image], axis=0)
            if len(image.shape) < idims:
                image = image[...,np.newaxis]
            if label is not None:
                ldims = len(label.shape)
                label = np.stack([cv2.resize(x, (targetsize[1], targetsize[0]), interpolation = cv2.INTER_NEAREST) for x in label], axis=0)
                if len(label.shape) < ldims:
                    label = label[...,np.newaxis]
        if label is not None:
            return image, label
        return image
        
    def initAugmentation(self):
        # imgaug was compared to albumentation and imgaug was faster (see test sheet)
        # problem with albumentation it only takes a single image as input and processes 1-by-1

        self.currentTile = 0
        gaussian = lambda aug: iaa.Sometimes(self.prob_gaussian, aug)
        piecewise = lambda aug: iaa.Sometimes(self.prob_piecewise, aug)
        dropout = lambda aug: iaa.Sometimes(self.prob_dropout, aug)
        affine = lambda aug: iaa.Sometimes(self.prob_affine, aug)
        

        self.augmentation_sequence = iaa.Sequential()
        
        if not self.enableAll:
            return
        
        if self.flip_horz:
            self.augmentation_sequence.append(iaa.Fliplr(0.5))
            
        if self.flip_vert:
            self.augmentation_sequence.append(iaa.Flipud(0.5))
            
        if self.prob_gaussian > 0:
            self.augmentation_sequence.append(gaussian(iaa.GaussianBlur(sigma=(0, 3.0))))
            
        if self.prob_piecewise > 0:
            self.augmentation_sequence.append(piecewise(iaa.PiecewiseAffine(scale=(0.01, 0.05))))
            
        if self.prob_dropout > 0:
            self.augmentation_sequence.append(dropout(iaa.OneOf([
                            iaa.Dropout((0.01, 0.1), per_channel=0.5),
                            iaa.CoarseDropout((0.03, 0.15), size_percent=(0.02, 0.05), per_channel=0.2),
                ])))
            
        if self.enableAffine and self.prob_affine > 0:
            self.augmentation_sequence.append(affine(
                    iaa.KeepSizeByResize(
                    iaa.Affine(
                    scale={"x": (self.scale_min, self.scale_max), "y": (self.scale_min, self.scale_max)},
                    translate_percent={"x": (self.translate_min, self.translate_max), "y": (self.translate_min, self.translate_max)},
                    rotate=(self.rotate_min, self.rotate_max), 
                    shear=(self.shear_min, self.shear_max), 
                    order=[0, 1], 
                    mode=["reflect"] # what is best padding strategy for segmentation ?!
                ))))

            
    def augment(self, image, label, validation = False):     
        if self.parent.Mode.type == dlMode.Segmentation or self.parent.Mode.type == dlMode.Object_Counting:
            if validation:
                x, y = self.systematiccrop_or_pad(image, label)
            else:
                image, label = self.randomcrop_or_pad(image, label)
                x,y = self.augmentation_sequence(images=image, segmentation_maps=label)
            x, y = self.checkModelSizeCompatibility(np.asarray(x), np.asarray(y))

        elif self.parent.Mode.type == dlMode.Classification:
            if validation:
                x = self.fitFullSizeImage2OutputSize(image)
                y = label
            else:
                image = self.fitFullSizeImage2OutputSize(image)
                y = label
                x = self.augmentation_sequence(images=image)
            
            x = self.checkModelSizeCompatibility(np.asarray(x))
            y = np.asarray(y)
        else:
            raise 
        return x,y




from dl.models.classification_models import get_submodules_from_kwargs


def slice_tensor(x, start, stop, axis):
    if axis == 3:
        return x[:, :, :, start:stop]
    elif axis == 1:
        return x[:, start:stop, :, :]
    else:
        raise ValueError("Slice axis should be in (1, 3), got {}.".format(axis))


def GroupConv2D(filters,
                kernel_size,
                strides=(1, 1),
                groups=32,
                kernel_initializer='he_uniform',
                use_bias=True,
                activation='linear',
                padding='valid',
                **kwargs):
    """
    Grouped Convolution Layer implemented as a Slice,
    Conv2D and Concatenate layers. Split filters to groups, apply Conv2D and concatenate back.

    Args:
        filters: Integer, the dimensionality of the output space
            (i.e. the number of output filters in the convolution).
        kernel_size: An integer or tuple/list of a single integer,
            specifying the length of the 1D convolution window.
        strides: An integer or tuple/list of a single integer, specifying the stride
            length of the convolution.
        groups: Integer, number of groups to split input filters to.
        kernel_initializer: Regularizer function applied to the kernel weights matrix.
        use_bias: Boolean, whether the layer uses a bias vector.
        activation: Activation function to use (see activations).
            If you don't specify anything, no activation is applied (ie. "linear" activation: a(x) = x).
        padding: one of "valid" or "same" (case-insensitive).

    Input shape:
        4D tensor with shape: (batch, rows, cols, channels) if data_format is "channels_last".

    Output shape:
        4D tensor with shape: (batch, new_rows, new_cols, filters) if data_format is "channels_last".
        rows and cols values might have changed due to padding.

    """

    backend, layers, models, keras_utils = get_submodules_from_kwargs(kwargs)
    slice_axis = 3 if backend.image_data_format() == 'channels_last' else 1

    def layer(input_tensor):
        inp_ch = int(backend.int_shape(input_tensor)[-1] // groups)  # input grouped channels
        out_ch = int(filters // groups)  # output grouped channels

        blocks = []
        for c in range(groups):
            slice_arguments = {
                'start': c * inp_ch,
                'stop': (c + 1) * inp_ch,
                'axis': slice_axis,
            }
            x = layers.Lambda(slice_tensor, arguments=slice_arguments)(input_tensor)
            x = layers.Conv2D(out_ch,
                              kernel_size,
                              strides=strides,
                              kernel_initializer=kernel_initializer,
                              use_bias=use_bias,
                              activation=activation,
                              padding=padding)(x)
            blocks.append(x)

        x = layers.Concatenate(axis=slice_axis)(blocks)
        return x

    return layer


def expand_dims(x, channels_axis):
    if channels_axis == 3:
        return x[:, None, None, :]
    elif channels_axis == 1:
        return x[:, :, None, None]
    else:
        raise ValueError("Slice axis should be in (1, 3), got {}.".format(channels_axis))


def ChannelSE(reduction=16, **kwargs):
    """
    Squeeze and Excitation block, reimplementation inspired by
        https://github.com/Cadene/pretrained-models.pytorch/blob/master/pretrainedmodels/models/senet.py

    Args:
        reduction: channels squeeze factor

    """
    backend, layers, models, keras_utils = get_submodules_from_kwargs(kwargs)
    channels_axis = 3 if backend.image_data_format() == 'channels_last' else 1

    def layer(input_tensor):
        # get number of channels/filters
        channels = backend.int_shape(input_tensor)[channels_axis]

        x = input_tensor

        # squeeze and excitation block in PyTorch style with
        x = layers.GlobalAveragePooling2D()(x)
        x = layers.Lambda(expand_dims, arguments={'channels_axis': channels_axis})(x)
        x = layers.Conv2D(channels // reduction, (1, 1), kernel_initializer='he_uniform')(x)
        x = layers.Activation('relu')(x)
        x = layers.Conv2D(channels, (1, 1), kernel_initializer='he_uniform')(x)
        x = layers.Activation('sigmoid')(x)

        # apply attention
        x = layers.Multiply()([input_tensor, x])

        return x

    return layer



class KerasObject:
    _backend = None
    _models = None
    _layers = None
    _utils = None

    def __init__(self, name=None):
        if (self.backend is None or
                self.utils is None or
                self.models is None or
                self.layers is None):
            raise RuntimeError('You cannot use `KerasObjects` with None submodules.')

        self._name = name

    @property
    def __name__(self):
        if self._name is None:
            return self.__class__.__name__
        return self._name

    @property
    def name(self):
        return self.__name__

    @name.setter
    def name(self, name):
        self._name = name

    @classmethod
    def set_submodules(cls, backend, layers, models, utils):
        cls._backend = backend
        cls._layers = layers
        cls._models = models
        cls._utils = utils

    @property
    def submodules(self):
        return {
            'backend': self.backend,
            'layers': self.layers,
            'models': self.models,
            'utils': self.utils,
        }

    @property
    def backend(self):
        return self._backend

    @property
    def layers(self):
        return self._layers

    @property
    def models(self):
        return self._models

    @property
    def utils(self):
        return self._utils


class Metric(KerasObject):
    pass


class Loss(KerasObject):

    def __add__(self, other):
        if isinstance(other, Loss):
            return SumOfLosses(self, other)
        else:
            raise ValueError('Loss should be inherited from `Loss` class')

    def __radd__(self, other):
        return self.__add__(other)

    def __mul__(self, value):
        if isinstance(value, (int, float)):
            return MultipliedLoss(self, value)
        else:
            raise ValueError('Loss should be inherited from `BaseLoss` class')

    def __rmul__(self, other):
        return self.__mul__(other)


class MultipliedLoss(Loss):

    def __init__(self, loss, multiplier):

        # resolve name
        if len(loss.__name__.split('+')) > 1:
            name = '{}({})'.format(multiplier, loss.__name__)
        else:
            name = '{}{}'.format(multiplier, loss.__name__)
        super().__init__(name=name)
        self.loss = loss
        self.multiplier = multiplier

    def __call__(self, gt, pr):
        return self.multiplier * self.loss(gt, pr)


class SumOfLosses(Loss):

    def __init__(self, l1, l2):
        name = '{}_plus_{}'.format(l1.__name__, l2.__name__)
        super().__init__(name=name)
        self.l1 = l1
        self.l2 = l2

    def __call__(self, gt, pr):
        return self.l1(gt, pr) + self.l2(gt, pr)



# -*- coding: utf-8 -*-
"""
Created on Thu Apr 16 14:56:23 2020

@author: Koerber
"""

import os
import glob
import cv2
from dl.data.labels import getMatchingImageLabelPairsRecursive
import numpy as np
from utils.Image import ImageFile
import random
import concurrent.futures
from itertools import repeat
import math
import random
import time


class ImageData():
    def __init__(self, parent):
        self.parent = parent
        
        # should we split validation data and image data in 2 dataset classes ?
        self.TrainingImagePaths = []
        self.TrainingLabelPaths = []
        self.TrainingTileIndices = []
        
        # TrainTestSplit = 1 means 100% Training Data and 0% validation
        self.TrainTestSplit = 0.85
        self.UseValidationDir = False
        
        self.IgnoreBorder = 0
        
        self.autocalcClassWeights = False
        self.class_weights = None 
        self.classValues = None
        
        self.ValidationImagePaths = []
        self.ValidationLabelPaths = []
        self.ValidationTileIndices = []

    def getImagePaths(self, validation):
        if validation:
            return self.ValidationImagePaths
        else:
            return self.TrainingImagePaths
            
        
    def getLabelPaths(self, validation):
        if validation:
            return self.ValidationLabelPaths
        else:
            return self.TrainingLabelPaths
    
            
    def addTileIndices(self, tiles, validation):
        if validation:
            self.ValidationTileIndices.extend(tiles)
        else:
            self.TrainingTileIndices.extend(tiles)
        
    def readImage(self, path, frame = 0):
        file = ImageFile(path)
        # should we normalize each image or over the dataset?
        # pro dataset: better for images from the same source, i.e. images that do not contain any target will lead to background that gets amplified by single image normalization
        # con dataset: single image normalization is better suited for different sources, like 8-bit and 16-bit and color images combined
        file.normalizeImage()
        image = file.getDLInputImage(self.parent.MonoChrome,frame)
        return image
    
    
    def initTrainingDataset(self, train_imagepath, val_imagepath):
        images, labels = getMatchingImageLabelPairsRecursive(train_imagepath, self.parent.LabelFolderName)
        if images is None or labels is None or not images or not labels:
            return None, None
        
        
             
        # careful: we got new validation set each time we start training
        z = list(zip(images, labels))

        random.shuffle(z)
        images, labels = zip(*z)


        if self.UseValidationDir:  
            self.TrainingImagePaths = images
            self.TrainingLabelPaths = labels
            self.TrainingTileIndices = []
            
            if val_imagepath and os.path.isdir(val_imagepath):
                val_images, val_labels = getMatchingImageLabelPairsRecursive(val_imagepath, self.parent.LabelFolderName)
                self.ValidationImagePaths = val_images
                self.ValidationLabelPaths = val_labels
                self.ValidationTileIndices = [] 
        else:
            split = int(self.TrainTestSplit*len(images))
    
            if split > 0:
                self.TrainingImagePaths = images[0:split]
                self.TrainingLabelPaths = labels[0:split]
                self.TrainingTileIndices = []
                self.ValidationImagePaths = images[split:]
                self.ValidationLabelPaths = labels[split:]
                self.ValidationTileIndices = []
            else:
                self.TrainingImagePaths = images
                self.TrainingLabelPaths = labels
                self.TrainingTileIndices = []



    def initialized(self):
        return True if len(self.TrainingImagePaths) * len(self.TrainingLabelPaths) > 0 else False
    
    def validationData(self):
        return True if len(self.ValidationImagePaths) * len(self.ValidationLabelPaths) > 0 else False

    def preprocessImage(self, image):
        return self.parent.Mode.preprocessImage(image)
        
    def preprocessLabel(self,label):
        return self.parent.Mode.preprocessLabel(label)
    
    
    def removeUnlabelledData(self, image, label):
        _, thresh = cv2.threshold(label, 0, 255,  cv2.THRESH_BINARY)
        image = cv2.bitwise_and(image,image, mask=thresh)
            
        # remove background
        label[np.where(label == 255)] = 0
        return image, label

            
    def createImageLabelPair(self, index, validation=False):
        images = self.getImagePaths(validation)
        masks = self.getLabelPaths(validation)

        channels = 1 if self.parent.MonoChrome is True else 3

        # handle image stacks and others
        # using same label for complete stack unhandled atm
        if isinstance(masks[index], tuple):
            mask = masks[index][0]
            try:
                frame = int(masks[index][1])
            except:
                print(masks[index][1])
                raise # not a valid label for a stack
            image = images[index][0]
        else:
            mask = masks[index]
            frame = None
            image = images[index]
            
        train_img = self.readImage(image, frame)

        width, height = self.parent.Mode.getImageSize4ModelInput(train_img.shape[1], train_img.shape[0])
              
        train_mask = self.parent.Mode.LoadLabel(mask, train_img.shape[0], train_img.shape[1])

        train_img, train_mask = self.removeUnlabelledData(train_img, train_mask) 
        train_img =  cv2.resize(train_img, (width, height))
        train_mask = self.parent.Mode.resizeLabel(train_mask, (width, height))

        if self.autocalcClassWeights and not validation:
            if self.classValues is not None:
                self.classValues += self.parent.Mode.countLabelWeight(train_mask)

        train_mask = train_mask.reshape(height, width, 1)
        train_img = train_img.reshape(height, width, channels)
        
        if self.IgnoreBorder > 0 and 3*self.IgnoreBorder < height and 3*self.IgnoreBorder < width:
            train_mask = train_mask[self.IgnoreBorder:height-2*self.IgnoreBorder,self.IgnoreBorder:width-2*self.IgnoreBorder,:].copy()
            train_img = train_img[self.IgnoreBorder:height-2*self.IgnoreBorder,self.IgnoreBorder:width-2*self.IgnoreBorder,:].copy()
        

        train_mask = self.parent.Mode.prepreprocessLabel(train_mask)
        return train_img, train_mask  
    
    def getTileIndices(self, validation = False, equalTilesperImage = False):
        if not equalTilesperImage:
            if validation:
                self.ValidationTileIndices.sort()
                return self.ValidationTileIndices
            else:
                return self.TrainingTileIndices
        else:
            images = self.getImagePaths(validation)
            numImages = len(images)
            # draw xtimes an images and calculate tilenumber and equally distribute
            xtimes = 5
            tilenum = 0
            for _ in range(xtimes):
                index = random.randint(0,numImages-1)
                
                if isinstance(images[index], tuple):
                    image = images[index][0]
                else:
                    image = images[index]
                img = self.readImage(image)
                width = int(img.shape[1]*self.parent.ImageScaleFactor)
                height= int(img.shape[0]*self.parent.ImageScaleFactor)
                tilenum += self.parent.augmentation.getTilesPerImage(width, height)
            tilenum = tilenum // xtimes
            ret = []
            [ret.extend([i] * tilenum) for i in range(numImages)]
            if validation:
                ret.sort()
                self.ValidationTileIndices = ret
            else:
                self.TrainingTileIndices = ret
            return ret
        
    def getNumberOfBatches(self, validation = False):
        if validation:
            idc = len(self.ValidationTileIndices)
        else:
            idc = len(self.TrainingTileIndices)
            
        return int(np.ceil(idc / float(self.parent.batch_size)))
    
    def loadTrainingDataSet(self, validation = False):
        if not self.initialized():
            return None, None
        images = self.getImagePaths(validation)
        numImages = len(images)
        if not validation:
            self.initclassValues()

        im_channels = 1 if self.parent.MonoChrome is True else 3
        with concurrent.futures.ThreadPoolExecutor(max_workers=self.parent.worker) as executor:
            x = executor.map(self.createImageLabelPair,range(numImages), repeat(validation))
        if not validation:
            self.calcClassWeights()
        img = []
        mask = []

        for counter, pair in enumerate(x): 
            tiles = [counter] * self.parent.augmentation.getTilesPerImage(pair[0].shape[1],pair[0].shape[0])
            self.addTileIndices(tiles, validation)
            img.append(pair[0])
            mask.append(pair[1])
        return img, mask

    def getAutoClassWeights(self):
        # this is only used when not training in memory
        # up to 1000 images of training set are loaded and classweights calculated
        # would be much faster to only load labels 
        if not self.initialized():
            return
        self.initclassValues()
        if self.autocalcClassWeights:
            numImages = len(self.getImagePaths(False))
            for i in range(min(1000,numImages)):
                self.createImageLabelPair(i)
                
        self.calcClassWeights()
        
    def setClassWeights(self, weights):
        self.class_weights = weights
        
    def resetClassWeights(self):
        self.class_weights = None

    def initclassValues(self):
        self.classValues = np.zeros((self.parent.NumClasses_real), dtype = np.float)

    def calcClassWeights(self):
        if self.autocalcClassWeights:
            self.class_weights = np.sum(self.classValues)/self.classValues
            self.class_weights /= np.sum(self.class_weights)
            if self.parent.NumClasses_real == 2:
                self.class_weights = self.class_weights[1]



# -*- coding: utf-8 -*-
"""
Created on Wed Aug 12 16:03:00 2020

@author: Koerber
"""

from tensorflow.keras.optimizers import Adam, Adamax, Adadelta, Adagrad, SGD
from enum import Enum
from inspect import signature

class dlOptim(Enum):
    adam = 1
    adamax = 2
    adadelta = 3
    adagrad = 4
    SGDNesterov = 5


class Optimizer():
    def __init__(self,parent):
        self.parent = parent
        self.Optimizer = dlOptim.adam
        
    def getOptimizer(self):
        if self.Optimizer == dlOptim.adam:
            return Adam(learning_rate=self.parent.learning_rate)
        if self.Optimizer == dlOptim.adamax:
            return Adamax(learning_rate=self.parent.learning_rate)
        if self.Optimizer == dlOptim.adadelta:
            return Adadelta(learning_rate=self.parent.learning_rate)
        if self.Optimizer == dlOptim.adagrad:
            return Adagrad(learning_rate=self.parent.learning_rate)
        if self.Optimizer == dlOptim.SGDNesterov:
            return SGD(learning_rate=self.parent.learning_rate, momentum=0.9, nesterov=True) 
        else:
            raise # not supported
    
    def setOptimizer(self, opti):
        self.Optimizer = opti
        if self.Optimizer == dlOptim.adam:
            fun = Adam
        elif self.Optimizer == dlOptim.adamax:
            fun = Adamax
        elif self.Optimizer == dlOptim.adadelta:
            fun = Adadelta
        elif self.Optimizer == dlOptim.adagrad:
            fun = Adagrad
        elif self.Optimizer == dlOptim.SGDNesterov:
            fun = SGD  
        else:
            raise # not supported
        self.parent.learning_rate = signature(fun).parameters['learning_rate'].default
        


""" Utility functions for segmentation models """

from dl.models.keras_applications import get_submodules_from_kwargs
from . import inject_global_submodules


def set_trainable(model, recompile=True, **kwargs):
    """Set all layers of model trainable and recompile it

    Note:
        Model is recompiled using same optimizer, loss and metrics::

            model.compile(
                model.optimizer,
                loss=model.loss,
                metrics=model.metrics,
                loss_weights=model.loss_weights,
                sample_weight_mode=model.sample_weight_mode,
                weighted_metrics=model.weighted_metrics,
            )

    Args:
        model (``keras.models.Model``): instance of keras model

    """
    for layer in model.layers:
        layer.trainable = True

    if recompile:
        model.compile(
            model.optimizer,
            loss=model.loss,
            metrics=model.metrics,
            loss_weights=model.loss_weights,
            sample_weight_mode=model.sample_weight_mode,
            weighted_metrics=model.weighted_metrics,
        )


@inject_global_submodules
def set_regularization(
        model,
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        beta_regularizer=None,
        gamma_regularizer=None,
        **kwargs
):
    """Set regularizers to all layers

    Note:
       Returned model's config is updated correctly

    Args:
        model (``keras.models.Model``): instance of keras model
        kernel_regularizer(``regularizer`): regularizer of kernels
        bias_regularizer(``regularizer``): regularizer of bias
        activity_regularizer(``regularizer``): regularizer of activity
        gamma_regularizer(``regularizer``): regularizer of gamma of BatchNormalization
        beta_regularizer(``regularizer``): regularizer of beta of BatchNormalization

    Return:
        out (``Model``): config updated model
    """
    _, _, models, _ = get_submodules_from_kwargs(kwargs)

    for layer in model.layers:
        # set kernel_regularizer
        if kernel_regularizer is not None and hasattr(layer, 'kernel_regularizer'):
            layer.kernel_regularizer = kernel_regularizer
        # set bias_regularizer
        if bias_regularizer is not None and hasattr(layer, 'bias_regularizer'):
            layer.bias_regularizer = bias_regularizer
        # set activity_regularizer
        if activity_regularizer is not None and hasattr(layer, 'activity_regularizer'):
            layer.activity_regularizer = activity_regularizer

        # set beta and gamma of BN layer
        if beta_regularizer is not None and hasattr(layer, 'beta_regularizer'):
            layer.beta_regularizer = beta_regularizer

        if gamma_regularizer is not None and hasattr(layer, 'gamma_regularizer'):
            layer.gamma_regularizer = gamma_regularizer

    out = models.model_from_json(model.to_json())
    out.set_weights(model.get_weights())

    return out



import keras
from .models_factory import ModelsFactory


class KerasModelsFactory(ModelsFactory):

    @staticmethod
    def get_kwargs():
        return {
            'backend': keras.backend,
            'layers': keras.layers,
            'models': keras.models,
            'utils': keras.utils,
        }


Classifiers = KerasModelsFactory()



from tensorflow.keras.callbacks import Callback
import csv
import tensorflow.keras.backend as K

class TrainingRecording(Callback):
    def __init__(self, parent ):
        self.parent = parent
        self.loss = []
        self.loss_per_epoch = [0.035]
        
        self.metric = []
        self.metric_per_epoch = [0.05]
        
        self.val_loss = [0.045]
        self.val_metric = [0.08]
        self.lr = []
        self.interrupt = False
        self.currentbatch = 0
        self.currentepoch = 0
        # to do: implement validation loss and metric
        
    def save(self, path):
        header = ['epoch', 'train_loss', 'train_metric','learning_rate', 'val_loss', 'val_metric'] 

        with open (path, 'w') as f:
            write = csv.writer(f)
            write.writerow(header)
            for i in range(len(self.loss_per_epoch)):
                row = []
                row.append(i)
                row.append(self.loss_per_epoch[i])
                row.append(self.metric_per_epoch[i])
                if len(self.lr) > i:
                        row.append(self.lr[i])
                if len(self.val_loss) > i:
                    row.append(self.val_loss[i])
                if len(self.val_loss) > i:
                    row.append(self.val_metric[i])
                
                write.writerow(row)
                

    def on_train_begin(self, logs=None):
        self.parent.notifyTrainingStarted()
        self.interrupt = False
        self.currentepoch = self.parent.resumeEpoch
        self.currentbatch = 0
        
    def on_train_end(self, logs=None):
        self.parent.notifyTrainingFinished()

    def on_batch_end(self, batch, logs=None):
        self.loss.append(logs.get('loss'))
        self.metric.append(logs.get(self.parent.Model.metrics_names[1]))
        self.currentbatch = batch
        self.parent.notifyTrainingProgress()
        if self.interrupt:
            self.model.stop_training = True

    def on_epoch_end(self, epoch, logs=None):
        if self.model.stop_training:
            return 
        self.currentepoch = epoch + 1
        self.parent.notifyEpochEnd(epoch)
        l = logs.get('val_loss')
        m = logs.get('val_' + self.parent.Model.metrics_names[1])
        lpe = logs.get('loss')
        mpe = logs.get(self.parent.Model.metrics_names[1])
        lr = K.eval(self.parent.Model.optimizer.lr)
        
        if l:
            self.val_loss.append(l)
        if m:
            self.val_metric.append(m)
        if lpe:
            self.loss_per_epoch.append(lpe)
        if mpe:
            self.metric_per_epoch.append(mpe)
        if lr:
            self.lr.append(lr)

    def reset(self):
        self.loss = []
        self.loss_per_epoch = [0.035]
        self.metric = []
        self.metric_per_epoch = [0.05]
        self.val_loss = [0.045]
        self.val_metric = [0.08]
        self.lr = []
        self.currentepoch = 0
        self.currentbatch = 0




#!/usr/bin/env python
import numpy as np
from scipy import misc
from tensorflow.keras import backend as K

import tensorflow as tf
import dl.machine_learning.dextr.resnet as resnet



class DEXTR(object):
    """Pyramid Scene Parsing Network by Hengshuang Zhao et al 2017"""

    def __init__(self, nb_classes, resnet_layers, input_shape, weights, num_input_channels=4,
                 classifier='psp', use_numpy=False, sigmoid=False):
        self.input_shape = input_shape
        self.num_input_channels = num_input_channels
        self.sigmoid = sigmoid
        self.model = resnet.build_network(nb_classes=nb_classes, resnet_layers=resnet_layers, num_input_channels=num_input_channels,
                                          input_shape=self.input_shape, classifier=classifier, sigmoid=self.sigmoid, output_size=self.input_shape)
        self.model.load_weights(weights)

    def predict(self, img):
        # Preprocess
        img = misc.imresize(img, self.input_shape)
        img = img.astype('float32')
        probs = self.feed_forward(img)
        return probs

    def feed_forward(self, data):
        assert data.shape == (self.input_shape[0], self.input_shape[1], self.num_input_channels)
        prediction = self.model.predict(np.expand_dims(data, 0))[0]
        return prediction






# -*- coding: utf-8 -*-
"""
Created on Mon Jan 20 13:42:08 2020

@author: Koerber
"""

from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *

import csv
import os
import glob
import concurrent.futures
from utils.shapes.Contour import loadContours, getContourMinIntensity,getContourMeanIntensity, getContourMaxIntensity
from utils.shapes.Point import loadPoints
from ui.Tools import canvasTool
from ui.style import styleForm
from ui.ui_utils import DCDButton
from utils.Image import ImageFile
from dl.method.mode import dlMode
import cv2


class Window(object):
    def setupUi(self, Form):
        Form.setWindowTitle('Segmentation Settings') 

        styleForm(Form)
        self.centralWidget = QWidget(Form)

        self.vlayout = QVBoxLayout(self.centralWidget)
        self.vlayout.setContentsMargins(0, 0, 0, 0)
        self.vlayout.setSpacing(6)
        self.centralWidget.setLayout(self.vlayout)
        
        
        self.CBSkeleton = QCheckBox("Export Skeleton",self.centralWidget)
        self.CBSkeleton.setObjectName('ExportSkeleton')
        self.CBSkeleton.setChecked(True)
        self.CBSkeleton.setToolTip('Check to export skeleton length')

        self.CBPerimeter = QCheckBox("Export Perimeter",self.centralWidget)
        self.CBPerimeter.setObjectName('Exportperimeter')
        self.CBPerimeter.setChecked(True)
        self.CBPerimeter.setToolTip('Check to export Perimeter length')

        self.CBMin = QCheckBox("Export Min Intensity",self.centralWidget)
        self.CBMin.setObjectName('ExportMin')
        self.CBMin.setChecked(True)
        self.CBMin.setToolTip('Check to export minimum pixel intensity')

        self.CBMean = QCheckBox("Export Mean Intensity",self.centralWidget)
        self.CBMean.setObjectName('ExportMean')
        self.CBMean.setChecked(True)
        self.CBMean.setToolTip('Check to export mean pixel intensity')

        self.CBMax = QCheckBox("Export Max Intensity",self.centralWidget)
        self.CBMax.setObjectName('ExportMax')
        self.CBMax.setChecked(True)
        self.CBMax.setToolTip('Check to export maximum pixel intensity')

        self.vlayout.addWidget(self.CBSkeleton)
        self.vlayout.addWidget(self.CBPerimeter)
        self.vlayout.addWidget(self.CBMin)
        self.vlayout.addWidget(self.CBMean)
        self.vlayout.addWidget(self.CBMax)

        self.centralWidget.setFixedSize(self.vlayout.sizeHint())
        Form.setFixedSize(self.vlayout.sizeHint())
                
        


class SegmentationResultsWindow(QMainWindow, Window):
    def __init__(self, parent, parentwindow):
        super(SegmentationResultsWindow, self).__init__(parent)
        self.parent = parent
        self.parentwindow = parentwindow
        self.setupUi(self)
        self.funcs = []


    def saveContourData(self, writer, images, labels):    
        self.funcs = []
        if self.parent.TrackingModeEnabled:
            self.saveTrackResults(writer, images, labels)
        else:
            self.saveSegmentationResults(writer, images, labels)

    def saveSegmentationResults(self,writer, images, labels):
        header = ['image name'] + ['frame number'] + ['object number'] + ['object type'] 

        header = self.addOptionalParameters(header)
                    
        writer.writerow(header)
        num = len(labels)
        self.parent.emitinitProgress(num)
        with concurrent.futures.ThreadPoolExecutor(max_workers=self.parent.maxworker) as executor:
            rows = executor.map(self.getSingleContourLabelFromList, zip(images,labels))
        
        for r in rows:
             writer.writerows(r) 


    def addOptionalParameters(self, header):
        f1 = lambda x, _: x.getSize()
        f2 = lambda x, _: x.getSize() /((self.parent.canvas.scale_pixel_per_mm/1000)**2)
        f3 = lambda x, _: x.getSkeletonLength(self.parent.canvas.skeletonsmoothingfactor) 
        f4 = lambda x, _: x.getSkeletonLength(self.parent.canvas.skeletonsmoothingfactor)/(self.parent.canvas.scale_pixel_per_mm/1000)
        f5 = lambda x, _: x.getSize()/max(1,x.getSkeletonLength(self.parent.canvas.skeletonsmoothingfactor))
        f6 = lambda x, _: x.getPerimeter()
        f7 = lambda x, _: x.getPerimeter()/(self.parent.canvas.scale_pixel_per_mm/1000)
        f8 = lambda x, y: getContourMinIntensity(y,x)
        f9 = lambda x, y: getContourMeanIntensity(y,x)
        f10 = lambda x, y: getContourMaxIntensity(y,x)

        if self.parentwindow.CBSize.isChecked():
            header += ['size in microns\u00b2']
            self.funcs += (f2,)
        else:
            header += ['size']
            self.funcs += (f1,)
        if self.CBSkeleton.isChecked():
            if self.parentwindow.CBSize.isChecked():
                header += ['length in microns']
                self.funcs += (f4,)
            else:
                header += ['length in pixel']
                self.funcs += (f3,)
            header += ['fatness (area / length)']
            self.funcs += (f5,)
        if self.CBPerimeter.isChecked(): 
            if self.parentwindow.CBSize.isChecked():
                header += ['perimeter in microns']
                self.funcs += (f7,)
            else:
                header += ['perimeter']
                self.funcs += (f6,)
        if self.CBMin.isChecked():
            header += ['minimum pixel intensity']
            self.funcs += (f8,)
        if self.CBMean.isChecked():
            header += ['mean pixel intensity']
            self.funcs += (f9,)
        if self.CBMax.isChecked():
            header += ['max pixel intensity']
            self.funcs += (f10,)

        return header

    def LoadImageIfNecessary(self, name, frame):
        if self.CBMin.isChecked() or self.CBMean.isChecked() or self.CBMax.isChecked(): 
            imagefile = ImageFile(self.parent.files.convertIfStackPath(name))
            image = imagefile.getImage(frame)
        else: 
            image = None
        return image

    def getFrameNumber(self, label):
        if self.parent.files.isStackLabel(label):
            frame = self.parent.files.getFrameNumber(label) + 1
        else:
            frame = 1
        return frame
        
    def getSingleContourLabelFromList(self, im_labels):

        name = self.parent.files.getFilenameFromPath(self.parent.files.convertIfStackPath(im_labels[0]),withfileextension=True)
        contours = loadContours(self.parent.files.convertIfStackPath(im_labels[1]))
        contours = [x for x in contours if x.isValid(self.parent.canvas.minContourSize)]
        frame = self.getFrameNumber(im_labels[1])
        x = 0
        rows = []
        for i,c in enumerate(contours):
            x += 1
            row = [name] + [frame] 
            row += [c.objectNumber] if c.objectNumber >= 0 else [i+1]
            row += [c.classlabel]
            image = self.LoadImageIfNecessary(im_labels[0],frame-1)
            row += [f(c,image) for f in self.funcs]
            rows.append(row)
        self.parent.emitProgress()
        return rows

    def saveTrackResults(self, writer, images, labels):
        header = ['object number'] + ['time point'] +  ['image name'] + ['frame number'] + ['object type'] + ['position'] 
        f1 = lambda x, _: x.classlabel
        f2 = lambda x, _: x.getCenter()
        self.funcs = (f1,f2)
        header = self.addOptionalParameters(header)

        writer.writerow(header)
        num = len(labels)
        self.parent.emitinitProgress(num)
        trackres = self.getTrackResults(images, labels)
        for r in trackres:
            writer.writerow(r) 
        
    def getTRackresults(self, images,labels):
        contours = []
        contours = [{} for x in range(self.parent.tracking.timepoints)]

        for i,l in zip(images,labels):
            name = self.parent.files.getFilenameFromPath(self.parent.files.convertIfStackPath(i),withfileextension=True)
            contour = loadContours(self.parent.files.convertIfStackPath(l))
            if self.parent.tracking.stackMode and self.parent.files.isStackLabel(l):
                tp = self.parent.files.getFrameNumber(l)
            else:
                tp = self.parent.tracking.getTimePointFromImageName(self.parent.files.convertIfStackPath(i))
            frame = self.getFrameNumber(l)

            image = self.LoadImageIfNecessary(i,frame-1)
            contours[tp] = {x.objectNumber: [name] + [frame] + [f(x,image) for f in self.funcs] for x in contour if x.isValid(self.parent.canvas.minContourSize)}
            self.parent.emitProgress()
        rows = []
        for n in self.parent.tracking.getNumbersOfTrackedObjects():
            for t in self.parent.tracking.getObjectOccurences(n):
                try:
                    row = [n] + [t+1] + contours[t][n]
                    rows.append(row)
                except:
                    pass
        return rows



# -*- coding: utf-8 -*-
"""
Created on Wed Aug 12 15:32:20 2020

@author: Koerber
"""

import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras.utils import get_custom_objects
from dl.metric.metric_functions import dice_coefficient, dice_coefficient_binary, dice_coefficient_weighted, dice_coefficient_binary_weighted


## classification
# see segmentation

## regression
def mean_squared_error(class_weights=None):
    if class_weights is None:
        class_weights = 1
        
    def _mean_squared_error(y_true, y_pred):
        mse = K.mean(K.square(y_pred - y_true)*class_weights, axis=-1)
        return mse
    return _mean_squared_error

def mean_absolute_error(class_weights=None):
    if class_weights is None:
        class_weights = 1

    def _mean_absolute_error(y_true, y_pred):
        mae = K.mean(K.abs(y_true - y_pred)*class_weights, axis=-1)
        return mae
    return _mean_absolute_error

def mean_squared_logarithmic_error(class_weights=None):
    if class_weights is None:
        class_weights = 1

    def _mean_squared_logarithmic_error(y_true, y_pred):
        # clip at 0 or (-1 + epsilon) ??
        y_pred = K.clip(y_pred, 0, None)
        y_true = K.clip(y_true, 0, None)
        msle = K.mean(K.square(K.log(y_true + 1) - K.log(y_pred + 1))*class_weights, axis=-1)
        return msle
    return _mean_squared_logarithmic_error


## segmentation
def focal_loss(usedistmap=False, class_weights = None, gamma=2):
    if class_weights is None:
        class_weights = .25
    # would be numerically more stable to use logits directly tf.nn.softmax_cross_entropy_with_logits and extend for focal and weights
    # similar as https://github.com/umbertogriffo/focal-loss-keras/blob/master/losses.py
    def _focal_loss(ytrue, ypred):
        if usedistmap:
            weightmap = K.expand_dims(ytrue[...,-1], axis=3)
            ytrue = ytrue[...,:-1]
        else: 
            weightmap = 1

        epsilon = K.epsilon()
        ypred = K.clip(ypred, epsilon, 1. - epsilon)
        return -K.mean(class_weights * K.pow(1. - ypred, gamma) * ytrue * K.log(ypred) * weightmap, axis=-1)
    return _focal_loss

def focal_loss_binary(usedistmap=False, class_weights = None, gamma=2):
    if class_weights is None:
        class_weights = .25
    # would be numerically more stable to use logits directly tf.nn.sigmoid_cross_entropy_with_logits and extend for focal and weights
    # similar as https://github.com/umbertogriffo/focal-loss-keras/blob/master/losses.py
    def _focal_loss_binary(y_true, y_pred):
        if usedistmap:
            weightmap = K.expand_dims(y_true[...,-1], axis=3)
            y_true = y_true[...,:-1]
        else:
            weightmap = 1

        y_pred = K.clip(y_pred, K.epsilon(), 1.0 - K.epsilon())
        p_t = tf.where(K.equal(y_true, 1), y_pred, 1 - y_pred)
        alpha_factor = K.ones_like(y_true) * class_weights
        alpha_t = tf.where(K.equal(y_true, 1), alpha_factor, 1 - alpha_factor)

        return -K.mean(K.sum(alpha_t * K.pow((1 - p_t), gamma) * K.log(p_t) * weightmap, axis=1))

    return _focal_loss_binary


def jaccard_distance_loss(usedistmap=False, class_weights=None, smooth=100):
    if class_weights is None:
        class_weights = 1
    # from https://gist.github.com/wassname/f1452b748efcbeb4cb9b1d059dce6f96
    def _jaccard_distance_loss(y_true, y_pred):
        if usedistmap:
            weightmap = K.expand_dims(y_true[...,-1], axis=3)
            y_true = y_true[...,:-1]
        else:
            weightmap = 1
            
        intersec = K.abs(y_true * y_pred) * weightmap
        intersection = K.sum(intersec, axis=-1)
        intersection_w = K.sum(intersec*class_weights, axis=-1)
        sum_ = K.sum((K.abs(y_true) + K.abs(y_pred))*weightmap, axis=-1)
        jac = (intersection_w + smooth) / (sum_ - intersection + smooth)
        return (1 - jac) * smooth
    return _jaccard_distance_loss


def jaccard_distance_loss_binary(usedistmap=False, class_weights=None, smooth=100):
    if class_weights is None:
        class_weights = .5

    def _jaccard_distance_loss_binary(y_true, y_pred):
        if usedistmap:
            weightmap = K.expand_dims(y_true[...,-1], axis=3)
            y_true = y_true[...,:-1]
        else:
            weightmap = 1
            
        # my personal interpretation of weighted binary jaccard,
        # not entirely sure if that makes sense in all cases -> check systematically
        intersec = K.abs(y_true * y_pred) * weightmap
        inter_nom = K.sum(intersec*class_weights, axis=-1)
        inter_denom = K.sum(intersec*(1-class_weights), axis=-1)
        sum_ = K.sum((class_weights*K.abs(y_true) + (1-class_weights)*K.abs(y_pred))*weightmap, axis=-1)
        jac = (inter_nom + smooth) / (sum_ - inter_denom + smooth)
        return (1 - jac) * smooth
    return _jaccard_distance_loss_binary


def kullback_Leibler_divergence_loss_weighted(usedistmap=False):   
    def _kullback_Leibler_divergence_loss_weighted(y_true, y_pred):
        # class_weights not supported atm
        if usedistmap:
            y_true = y_true[...,:-1]

        y_true = K.clip(y_true, K.epsilon(), 1)
        y_pred = K.clip(y_pred, K.epsilon(), 1)
        return K.sum(y_true * K.log(y_true / y_pred), axis=-1)
    
        # equivalent to             
        # kld = tf.keras.losses.KLDivergence()
        # return kld(y_true, y_pred)
    
    return _kullback_Leibler_divergence_loss_weighted


# for loading models trained with custom loss
get_custom_objects()['_mean_squared_error'] = mean_squared_error()
get_custom_objects()['_mean_absolute_error'] = mean_absolute_error()
get_custom_objects()['_mean_squared_logarithmic_error'] = mean_squared_logarithmic_error()

get_custom_objects()['_focal_loss'] = focal_loss()
get_custom_objects()['_focal_loss_binary'] = focal_loss_binary()
get_custom_objects()['_jaccard_distance_loss'] = jaccard_distance_loss()
get_custom_objects()['_kullback_Leibler_divergence_loss_weighted'] = kullback_Leibler_divergence_loss_weighted()




# -*- coding: utf-8 -*-
"""
Created on Fri Jun 26 15:55:22 2020

@author: Koerber
"""



from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *

import os
import numpy as np
import cv2
from ui.painter.Painter import Painter
import utils.shapes.Contour as Contour
from ui.Tools import canvasTool
import copy 

class ContourPainter(Painter):
    def __init__(self, canvas):
        super(ContourPainter,self).__init__(canvas)
        self.tools.append(canvasTool.assign)
        self.tools.append(canvasTool.assist)
        self.tools.append(canvasTool.objectnumber)
        self.tools.append(canvasTool.objectcolor)
        self.tools.append(canvasTool.deleteobject)
        
        # self.mask = None
        self.smartmode = False

        
    @property
    def sketch(self):
        if self.smartmode:
            return self.mask
        else:
            return self.contoursketch
        
    def addShapes(self,shapes):
        self.contours.addShapes(shapes)

    def draw(self):
        super(ContourPainter, self).draw()

        # we can not use a qpainterpath for all contours 
        # as the order of drawing is important 
        # and color has to be the same for the path
        # update: additionally, path.subtracted does not work for multiple objects
        for c in self.shapes.shapes:  
            if c.classlabel != 0:
                self.drawcontour(c)             

    def enableDrawBackgroundMode(self):
        super(ContourPainter, self).enableDrawBackgroundMode()

    def disableDrawBackgroundMode(self):
        super(ContourPainter, self).enableDrawBackgroundMode()
        self.tools.append(canvasTool.assign)
        self.tools.append(canvasTool.assist)
        self.tools.append(canvasTool.objectnumber)
        self.tools.append(canvasTool.objectcolor)
        self.tools.append(canvasTool.deleteobject)

    def clear(self):
        super(ContourPainter, self).clear()
        self.shapes.clear()

    def load(self):
        self.clear()
        if self.canvas.parent.files.CurrentLabelPath() is not None and os.path.exists(self.canvas.parent.files.CurrentLabelPath()):
            self.shapes.load(self.canvas.parent.files.CurrentLabelPath())
        
    def save(self):
        if len(self.trackChanges) > self.maxChanges:
            self.trackChanges.pop(0)
        self.trackChanges.append(copy.deepcopy(self.shapes.shapes))
        self.canvas.parent.checkIfUndoPossible()
        if not self.shapes.empty():
            self.canvas.parent.ensureLabelFolder()
            self.shapes.save(self.canvas.parent.files.CurrentLabelPath())
        else:
            self.canvas.parent.deleteLabel()

    def undo(self):
        if len(self.trackChanges) < 2:
            return
        self.clear()
        contours = self.trackChanges[-2]
        self.trackChanges.pop()
        self.trackChanges.pop()
        self.shapes.addShapes(contours)
        self.canvas.parent.checkIfUndoPossible()
            
    def clearFOV(self, currentclass = False):
        width = self.canvas.image().width()
        height = self.canvas.image().height()
        self.contoursketch = np.zeros((height, width), np.uint8) 
        Contour.drawContoursToLabel(self.contoursketch, self.contours.shapes, drawbackground=False)
        fov = self.canvas.getFieldOfViewRect()
        if currentclass:
            rect = self.contoursketch[int(fov.y()):int(fov.y()+fov.height()),int(fov.x()):int(fov.x()+fov.width())]
            rect[rect == self.canvas.parent.activeClass()] = 0
            self.contoursketch[int(fov.y()):int(fov.y()+fov.height()),int(fov.x()):int(fov.x()+fov.width())] = rect
        else:
            self.contoursketch[int(fov.y()):int(fov.y()+fov.height()),int(fov.x()):int(fov.x()+fov.width())] = 0    
        contours = Contour.extractContoursFromLabel(self.contoursketch, not self.canvas.parent.allowInnerContours)
        self.contours.clear()
        self.contours.addShapes(contours)
        self.canvas.redrawImage()
        
    def prepareNewContour(self):
        if self.canvas.hasImage():
            width = self.canvas.image().width()
            height = self.canvas.image().height()

            if self.smartmode:
                if self.mask is None:
                    self.mask = np.zeros((height, width), np.uint8) + 2

            super(ContourPainter, self).prepareNewContour()


    def getFinalContours(self):
        if self.sketch is None:
            return
        
        if self.smartmode:
            self.clearFOV(currentclass = True)
            image = self.canvas.getFieldOfViewImage()
            fov = self.canvas.getFieldOfViewRect()
            mask = self.canvas.getFieldOfViewImage(self.mask)
            prediction = self.canvas.parent.dl.SemiAutoSegment(image, mask)
            self.contours.deleteShapes(self.contours.getShapesOfClass_x(self.canvas.parent.activeClass()))
            smartcontours = Contour.extractContoursFromImage(prediction, not self.canvas.parent.allowInnerContours, offset = (int(fov.x()),int(fov.y())))
            Contour.drawContoursToImage(self.contoursketch, smartcontours)  

        super(ContourPainter, self).getFinalContours()

    def assistedSegmentation(self, extremepoints):
        with self.canvas.parent.wait_cursor():
            # do not use fov here because the target image is cropped to network size
            prediction = self.canvas.parent.dl.AssistedSegmentation(self.canvas.parent.getCurrentImage(), np.asarray(extremepoints, dtype = np.int))
            if prediction is None:
                self.canvas.redrawImage()
                return
            shapes = Contour.extractContoursFromLabel(prediction, not self.canvas.parent.allowInnerContours)
            for s in shapes:
                s.setClassLabel(self.canvas.parent.activeClass())
            self.contours.addShapes(shapes)
            self.checkForChanges()
        
    def autosegment(self):
        if self.canvas.hasImage():
            self.clearFOV(currentclass = True)
            image = self.canvas.getFieldOfViewImage()
            fov = self.canvas.getFieldOfViewRect()
            # not using self.dl.dataloader.readImage() here as a 8-bit rgb image is required and no extra preprocessing
    
            if image is None:
                self.canvas.parent.PopupWarning('No image loaded')
                return
            try:
                prediction = self.canvas.parent.dl.AutoSegment(image)
            except:
                self.canvas.parent.PopupWarning('Auto prediction not possible')
                return
            shapes = Contour.extractContoursFromLabel(prediction, not self.canvas.parent.allowInnerContours, offset = (int(fov.x()),int(fov.y())))
            for s in shapes:
                s.setClassLabel(self.canvas.parent.activeClass())
            self.contours.addShapes(shapes)
            self.checkForChanges()
        
    def checkForChanges(self):
        if self.canvas.hasImage():
            width = self.canvas.image().width()
            height = self.canvas.image().height()
            self.contoursketch = np.zeros((height, width), np.uint8) 
            Contour.drawContoursToLabel(self.contoursketch, self.contours.shapes)
            
            contours = Contour.extractContoursFromLabel(self.contoursketch, not self.canvas.parent.allowInnerContours)
            # delete changed contours
            old_contours = self.contours.shapes
            changedcontours = Contour.getContoursNotinListOfContours(old_contours,contours)
            self.contours.deleteShapes(changedcontours)
                
            self.contours.addShapes(contours)
            self.canvas.redrawImage()
        


# -*- coding: utf-8 -*-
"""
Created on Wed Aug 12 15:52:38 2020

@author: Koerber
"""


from dl.metric.metrics import Metrics, dlMetric
from dl.metric.metric_functions import *

class RegressionMetrics(Metrics):
    def __init__(self,parent):
        super(RegressionMetrics,self).__init__(parent)
        self.supportedmetrics.append(dlMetric.mean_squared_err)
        self.supportedmetrics.append(dlMetric.root_mean_squared_err)
        self.supportedmetrics.append(dlMetric.mean_abs_err)
        self.supportedmetrics.append(dlMetric.mean_squared_log_err)
        self.metric = dlMetric.mean_abs_err
  
    def getMetric(self):
        if self.metric == dlMetric.mean_squared_err:
            return [mean_squared_error()]
        if self.metric == dlMetric.mean_abs_err:
            return [mean_absolute_error()]
        if self.metric == dlMetric.mean_squared_log_err:
            return [mean_squared_logarithmic_error()]
        if self.metric == dlMetric.root_mean_squared_err:
            return [root_mean_squared_error()]
        else:
            raise # not supported
            


from dl.models.keras_applications import get_submodules_from_kwargs


def Conv2dBn(
        filters,
        kernel_size,
        strides=(1, 1),
        padding='valid',
        data_format=None,
        dilation_rate=(1, 1),
        activation=None,
        kernel_initializer='glorot_uniform',
        bias_initializer='zeros',
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        use_batchnorm=False,
        **kwargs
):
    """Extension of Conv2D layer with batchnorm"""

    conv_name, act_name, bn_name = None, None, None
    block_name = kwargs.pop('name', None)
    backend, layers, models, keras_utils = get_submodules_from_kwargs(kwargs)

    if block_name is not None:
        conv_name = block_name + '_conv'

    if block_name is not None and activation is not None:
        act_str = activation.__name__ if callable(activation) else str(activation)
        act_name = block_name + '_' + act_str

    if block_name is not None and use_batchnorm:
        bn_name = block_name + '_bn'

    bn_axis = 3 if backend.image_data_format() == 'channels_last' else 1

    def wrapper(input_tensor):

        x = layers.Conv2D(
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilation_rate=dilation_rate,
            activation=None,
            use_bias=not (use_batchnorm),
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            name=conv_name,
        )(input_tensor)

        if use_batchnorm:
            x = layers.BatchNormalization(axis=bn_axis, name=bn_name)(x)

        if activation:
            x = layers.Activation(activation, name=act_name)(x)

        return x

    return wrapper



# -*- coding: utf-8 -*-
"""
Created on Thu Oct 10 13:28:08 2019

@author: Koerber
"""

from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
import sys
from tensorflow import errors

import io
import traceback
import glob
import os
import cv2 
import multiprocessing
import threading
import concurrent.futures
import time
from contextlib import contextmanager

import dl.DeepLearning as DeepLearning
from dl.method.mode import dlMode
from ui.UI import MainWindow
from ui.Tools import canvasTool, canvasToolButton
from ui.ui_Canvas import Canvas

from ui.segmentation.ui_SegPostProcessing import SegmentationPostProcessingWindow
from ui.objectdetection.ui_ODPostProcessing import ObjectDetectionPostProcessingWindow

from ui.ui_Results import ResultsWindow

from ui.ui_Training import TrainingWindow
from ui.ui_Settings import SettingsWindow

from ui.ui_TrainPlot import TrainPlotWindow
from ui.settings import Settings

import utils.shapes.Contour as Contour
import utils.shapes.Point as Point
from utils.Image import ImageFile, supportedImageFormats
from utils.Observer import QtObserver
from utils.FilesAndFolders import FilesAndFolders
from utils.postprocessing.tracking import ObjectTracking

from dl.data.labels import getMatchingImageLabelPairPaths

from utils.workerthread import WorkerThread
from ui.ui_utils import DCDButton, openFolder, saveFile, loadFile

import pickle

import numpy as np

LOG_FILENAME = 'log/logfile.log'
SETTINGS_FILENAME = 'settings/user_settings.json'



class MIA_UI(QMainWindow, MainWindow):
    # qt elements may only be changed from main thread, so we create signals here
    # to call from other threads
    initProgress = pyqtSignal(int)
    Progress = pyqtSignal() 
    ProgressFinished = pyqtSignal()
    Block = pyqtSignal() 
    UnBlock = pyqtSignal()
    Popup = pyqtSignal(str)
    StatusMessage = pyqtSignal(str)
    
    def __init__(self):
        super(MIA_UI, self).__init__()
        self.initialized = False
        self.setupUi(self)
        self.setCallbacks()
        self.show()

        self.progresstick = 10
        self.progress = 0
        self.progresslock = threading.Lock()
        self.processlock = threading.Lock()
        self.maxworker = 1
        self.lock = threading.Lock()
        
        self.files = FilesAndFolders(self)
        self.files.ensureFolder('log/')
        self.files.ensureFolder('settings/')
        
        self.currentImageFile = None
        self.keepImageSettings = False

        self.separateStackLabels = True
        self.setFocusPolicy(Qt.NoFocus)
        width = self.canvas.geometry().width()
        height = self.canvas.geometry().height() 
        
        
        self.predictionRate = 0
        self.saveload_modelweights_only = False
        
        self.dl = DeepLearning.DeepLearning()
        self.tracking = ObjectTracking(self.dl, self.files)
        # init observer
        self.dlobserver = QtObserver()
        self.dl.attachObserver(self.dlobserver)
        self.dlobserver.signals.progress.connect(self.dlProgress)
        self.dlobserver.signals.epoch_end.connect(self.dlEpoch)
        self.dlobserver.signals.finished.connect(self.dlFinished)
        self.dlobserver.signals.result.connect(self.dlResult)
        self.dlobserver.signals.error.connect(self.dlError)
        self.dlobserver.signals.start.connect(self.dlStarted)
        
        
        self.initProgress.connect(self.initializeProgress)
        self.Progress.connect(self.addProgress)
        self.ProgressFinished.connect(self.finishProgress)
        self.Block.connect(self.blockWindow)
        self.UnBlock.connect(self.unblockWindow)
        self.Popup.connect(self.PopupWarning)
        self.StatusMessage.connect(self.writeStatus)
        
        # init canvas
        self.hlayout.removeWidget(self.canvas)
        self.canvas.close()
        self.canvas = Canvas(self)
        self.canvas.setMouseTracking(True)
        self.canvas.setFocusPolicy(Qt.StrongFocus)
        self.canvas.setFixedSize(QSize(width, height))
        self.canvas.reset()
        self.hlayout.addWidget(self.canvas)
        self.hlayout.addWidget(self.SFrame)
        horizontalspacer = QSpacerItem(20, 40, QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.hlayout.addItem(horizontalspacer)
        
        self.classList.setClass(0)
        self.changeLearningMode(self.LearningMode().value)
        self.CBLearningMode.setCurrentIndex (self.LearningMode().value)
        self.setWorkers(multiprocessing.cpu_count() // 2)

        ### windows - should this be moved to UI ?
        self.segpostprocessing_form = SegmentationPostProcessingWindow(self)
        self.results_form = ResultsWindow(self)
        self.odpostprocessing_form = ObjectDetectionPostProcessingWindow(self)
        
        self.training_form = TrainingWindow(self)
        self.settings_form = SettingsWindow(self)
        
        self.plotting_form = TrainPlotWindow(self)
        
        self.ClassListUpdated()
        self.classList.setClass(1)
        
        
        self.initialized = True
        self.settings = Settings(self)
        self.settings.loadSettings(SETTINGS_FILENAME)
        
        self.defineShortcuts()
        


    def setCallbacks(self):
        self.Bclear.clicked.connect(self.clear)
        self.Bundo.clicked.connect(self.undo)
        self.BLoadTrainImages.clicked.connect(self.setTrainFolder)
        self.BLoadTestImages.clicked.connect(self.setTestFolder)
        self.BTestImageFolder.clicked.connect(self.switchToTestFolder)
        self.BTrainImageFolder.clicked.connect(self.switchToTrainFolder)
        self.Bnext.clicked.connect(self.nextImage)
        self.Bprev.clicked.connect(self.previousImage)
        self.SFrame.valueChanged.connect(self.FrameChanged)
        
        self.Bdrag.clicked.connect(self.setCanvasMode)
        self.Bdraw.clicked.connect(self.setCanvasMode)
        self.Bassign.clicked.connect(self.setCanvasMode)
        self.Bextend.clicked.connect(self.setCanvasMode)
        self.Bdelete.clicked.connect(self.setCanvasMode)
        self.Bpoly.clicked.connect(self.setCanvasMode)
        self.Bassist.clicked.connect(self.setCanvasMode)
        
        self.BassignClass.clicked.connect(self.setCanvasMode)
        self.BsetClass.clicked.connect(self.setCanvasMode)
        
        self.BsetObject.clicked.connect(self.setCanvasMode)
        self.BshiftObject.clicked.connect(self.setCanvasMode)

        self.BSetObjectNumber.clicked.connect(self.setCanvasMode)
        self.BDeleteObject.clicked.connect(self.setCanvasMode)
        self.BChangeObjectColor.clicked.connect(self.setCanvasMode)

        
        self.Btrain.clicked.connect(self.showTrainingWindow)
        self.Bpredictall.clicked.connect(self.predictAllImages)
        self.Bpredict.clicked.connect(self.predictImage)
        self.Bloadmodel.clicked.connect(self.loadModel)
        self.Bsavemodel.clicked.connect(self.saveModel)
        self.Bautoseg.clicked.connect(self.autoSegment)
        self.CBSmartMode.stateChanged.connect(self.setSmartMode)
        self.CBInner.stateChanged.connect(self.updateImage)

        self.Bresetmodel.clicked.connect(self.resetModel)
        self.Bautodect.clicked.connect(self.autoPointDetection)
        
        
        self.Bresults.clicked.connect(self.showResultsWindow)
        self.BSettings.clicked.connect(self.showSettingsWindow)
        self.BPostProcessing.clicked.connect(self.showPostProcessingWindow)
        
        self.Baddclass.clicked.connect(self.addClass)
        self.Bdelclass.clicked.connect(self.removeLastClass)

        self.BTrack.clicked.connect(self.calcTracking)
        self.CBshowtrack.stateChanged.connect(self.updateImage)
        self.CBtrackcolor.stateChanged.connect(self.updateImage)
        

        self.CBLearningMode.currentIndexChanged.connect(self.changeLearningMode)
        self.CBExtend.clicked.connect(self.updateCursor)
        self.CBErase.clicked.connect(self.updateCursor)
        self.CBAddShape.clicked.connect(self.setCanvasFocus)
        self.CBDelShape.clicked.connect(self.setCanvasFocus)
        self.CBSliceShape.clicked.connect(self.setCanvasFocus)
        
        self.SBrightness.sliderReleased.connect(self.setBrightness)
        self.SContrast.sliderReleased.connect(self.setContrast)
        self.CBKeep.stateChanged.connect(self.keepImageSettingsChanged)

        self.TVFiles.clicked.connect(self.FileExplorerItemSelected)
      
        
        ## menu actions
        self.ASetTrainingFolder.triggered.connect(self.setTrainFolder)
        self.ASetTestFolder.triggered.connect(self.setTestFolder)
        self.ASaveSettings.triggered.connect(self.SaveSettings)
        self.ALoadSettings.triggered.connect(self.LoadSettings)
        self.ASaveDeepLearning.triggered.connect(self.saveDLClass)
        
        self.AExit.triggered.connect(self.close)
        self.ASettings.triggered.connect(self.showSettingsWindow)
        
        # monkey patching download request for model weights
        import dl.utils.dl_downloads as down
        down.callBack = lambda: self.ConfirmPopup('Do you want to download model weights')
        
        
    def defineShortcuts(self):
        self.sc_toggle = QShortcut(QKeySequence(Qt.Key_Space), self)
        self.sc_toggle.activated.connect(self.canvas.toggleDrag) 
        
        self.sc_prev = QShortcut(QKeySequence(Qt.Key_A), self)
        self.sc_prev.activated.connect(self.previousImage) 
        
        self.sc_next = QShortcut(QKeySequence(Qt.Key_D), self)
        self.sc_next.activated.connect(self.nextImage) 
        
        self.sc_class0 = QShortcut(48, self)
        self.sc_class0.activated.connect(self.setClassShortcut) 
        
        self.sc_class1 = QShortcut(49, self)
        self.sc_class1.activated.connect(self.setClassShortcut) 
        
        self.sc_class2 = QShortcut(50, self)
        self.sc_class2.activated.connect(self.setClassShortcut) 
        
        self.sc_class3 = QShortcut(51, self)
        self.sc_class3.activated.connect(self.setClassShortcut)
        
        self.sc_class4 = QShortcut(52, self)
        self.sc_class4.activated.connect(self.setClassShortcut) 
        
        self.sc_class5 = QShortcut(53, self)
        self.sc_class5.activated.connect(self.setClassShortcut) 
        
        self.sc_class6 = QShortcut(54, self)
        self.sc_class6.activated.connect(self.setClassShortcut) 
        
        self.sc_class7 = QShortcut(55, self)
        self.sc_class7.activated.connect(self.setClassShortcut)
        
        self.sc_class8 = QShortcut(56, self)
        self.sc_class8.activated.connect(self.setClassShortcut) 
        
        self.sc_class9 = QShortcut(57, self)
        self.sc_class9.activated.connect(self.setClassShortcut) 
        
        self.sc_undo = QShortcut(QKeySequence(Qt.CTRL + Qt.Key_Z), self)
        self.sc_undo.activated.connect(self.undo) 

        self.sc_zoom_in = QShortcut(QKeySequence(Qt.Key_Plus), self)
        self.sc_zoom_in.activated.connect(self.canvas.zoomIn) 
        
        self.sc_zoom_out = QShortcut(QKeySequence(Qt.Key_Minus), self)
        self.sc_zoom_out.activated.connect(self.canvas.zoomOut) 
        
        self.sc_tool1 = QShortcut(QKeySequence(Qt.Key_F1), self)
        self.sc_tool1.activated.connect(self.setToolShortcut)
        
        self.sc_tool2 = QShortcut(QKeySequence(Qt.Key_F2), self)
        self.sc_tool2.activated.connect(self.setToolShortcut)
        
        self.sc_tool3 = QShortcut(QKeySequence(Qt.Key_F3), self)
        self.sc_tool3.activated.connect(self.setToolShortcut)
        
        self.sc_tool4 = QShortcut(QKeySequence(Qt.Key_F4), self)
        self.sc_tool4.activated.connect(self.setToolShortcut)
        
        self.sc_tool5 = QShortcut(QKeySequence(Qt.Key_F5), self)
        self.sc_tool5.activated.connect(self.setToolShortcut)
        
        self.sc_tool6 = QShortcut(QKeySequence(Qt.Key_F6), self)
        self.sc_tool6.activated.connect(self.setToolShortcut)

            
    def setClassShortcut(self):
        for i in range(10):
            val = i+48    
            if self.sender().key() == val:
                self.classList.setClass(i)
                break

    def setToolShortcut(self):
        if self.sender().key() == QKeySequence(Qt.Key_F1):
            self.selectToolButtonX(1)
        if self.sender().key() == QKeySequence(Qt.Key_F2):
            self.selectToolButtonX(2)
        if self.sender().key() == QKeySequence(Qt.Key_F3):
            self.selectToolButtonX(3)
        if self.sender().key() == QKeySequence(Qt.Key_F4):
            self.selectToolButtonX(4)
        if self.sender().key() == QKeySequence(Qt.Key_F5):
            self.selectToolButtonX(5)
        if self.sender().key() == QKeySequence(Qt.Key_F6):
            self.selectToolButtonX(6)


    def showSettingsWindow(self):
        self.settings_form.show()
        
    def showPostProcessingWindow(self):   
        if self.LearningMode() == dlMode.Segmentation:
            self.segpostprocessing_form.show()
        elif self.LearningMode() == dlMode.Object_Counting:
            self.odpostprocessing_form.show() 
    
    def closeEvent(self, event):       
        self.settings.saveSettings(SETTINGS_FILENAME)
        QApplication.closeAllWindows()
        self.dl.interrupt()
        super(QMainWindow, self).closeEvent(event)
        
    def SaveSettings(self):
        # filename = QFileDialog.getSaveFileName(self, "Save Settings", '', "Settings File (*.json)")[0] #<- native version
        filename = saveFile("Save Settings","Settings File (*.json)", 'json')
        if filename:
            self.settings.saveSettings(filename)
        
    def LoadSettings(self):
        # filename = QFileDialog.getOpenFileName(self, "Select Settings File", '',"Settings (*.json)")[0] #<- native version
        filename = loadFile("Select Settings File", "Settings (*.json)")
        if filename:
            self.settings.loadSettings(filename)
            
    def settingsLoaded(self):
        self.ClassListUpdated()

    def closeModeWindows(self):
        if self.initialized:
            self.odpostprocessing_form.hide()
            self.segpostprocessing_form.hide()

        
    def showResultsWindow(self):
        if self.files.testImageLabelspath is None:
            self.PopupWarning('Prediction folder not selected')
            return
        
        self.results_form.show()

        
    def PopupWarning(self, message):
        msg = QMessageBox(self)
        msg.setIcon(QMessageBox.Warning)
        msg.setText(message)
        msg.setWindowTitle("Warning")
        msg.setStandardButtons(QMessageBox.Ok)
        msg.exec_()  
        
    def ConfirmPopup(self, message):
        msg = QMessageBox(self)
        msg.setIcon(QMessageBox.Warning)
        msg.setText(message)
        msg.setWindowTitle("Confirmation")
        msg.setStandardButtons(QMessageBox.Ok | QMessageBox.Cancel)
        return msg.exec_() == QMessageBox.Ok
        
    def ShapesChanged(self):
        for i in range (self.NumOfClasses()):
            label = self.findChild(QLabel, 'numOfContours_class' + str(i))
            if i == 0:
                text = str(len(self.canvas.painter.backgroundshapes))
            elif self.canvas.painter.shapes:
                text = str(len(self.canvas.painter.shapes.getShapesOfClass_x(i)))
            else:
                text = ""
            label.setText(text)
        if self.TrackingModeEnabled:
            self.tracking.changeTimePoint(self.canvas.painter.shapes, self.files.currentFrame if self.tracking.stackMode else self.files.currentImage)
        self.canvas.SaveCurrentLabel()
        
    def setSmartMode(self):
        assert (self.LearningMode() == dlMode.Segmentation)
        self.canvas.painter.smartmode = self.CBSmartMode.isChecked()
        self.setCanvasFocus()
        
    @property
    def allowInnerContours(self):
        return self.CBInner.isChecked()
        
    def changeLearningMode(self, i):
        self.dl.WorkingMode = dlMode(i)
        self.canvas.setCanvasMode(self.LearningMode())
        self.closeModeWindows()
        if self.LearningMode() == dlMode.Segmentation:
            self.SegmentationSettings.show()
        else:
            self.SegmentationSettings.hide()
            
        if self.LearningMode() == dlMode.Object_Counting:
            self.DetectionSettings.show()
        else:
            self.DetectionSettings.hide()
        
        if self.LearningMode() == dlMode.Classification:
            self.ClassificationToolButtons.show()
        else:
            self.ClassificationToolButtons.hide()

        if self.initialized:
            # could be done in an observer
            self.results_form.ModeChanged()
            self.training_form.ModeChanged()
            if not self.TrackingSupported():
                self.enableTrackingMode(False)

            self.segpostprocessing_form.enableTrackingMode()
            self.odpostprocessing_form.enableTrackingMode()
            

        self.setWorkingFolder()

    def TrackingSupported(self):
        if self.LearningMode() == dlMode.Segmentation:
            return True
        if self.LearningMode() == dlMode.Object_Counting:
            return True
        else:
            return False

    def loadTrack(self):
        self.Tracking.setEnabled(False)
        if not self.files.testImagespath or not self.files.testImageLabelspath:
            return
        if self.TrackingModeEnabled and not self.files.train_test_dir:
            self.tracking.loadTrack()
            self.Tracking.setEnabled(True)
        
    def calcTracking(self):
        if not self.TrackingModeEnabled or not self.files.testImagespath or not self.files.testImageLabelspath:
            self.PopupWarning('No predicted files')
            return
        with self.wait_cursor():
            self.tracking.performTracking()
            self.canvas.ReloadImage()

    def enableTrackingMode(self, enable):
        if self.TrackingSupported():
            self.Tracking.setVisible(enable)
            self.loadTrack()
            self.canvas.ReloadLabels()
        else:
            self.Tracking.setVisible(False)
        
    def drawObjectTrack(self):
        return self.TrackingModeEnabled and self.CBshowtrack.isChecked()
            
    def colorCodeObjects(self):
        return self.TrackingModeEnabled and self.CBtrackcolor.isChecked()
    
    @property
    def TrackingModeEnabled(self):
        if not self.files.train_test_dir and self.TrackingSupported():
            return self.Tracking.isVisible()
        return False

    def setToolButtons(self):
        if self.activeClass() == 0:
            self.canvas.painter.enableDrawBackgroundMode()                
        else:
            self.canvas.painter.disableDrawBackgroundMode()
            
        # we need to first set all to invisible as otherwise the widget_width is extended 
        # and somewhat hard to control with pyqt :(
        for tool in canvasToolButton:
            widget = self.findChild(DCDButton, tool.name)
            if not widget:
                continue
            widget.setVisible(False)
        for tool in self.canvas.painter.tools: 
            widget = self.findChild(DCDButton, tool.name)
            if not widget:
                continue
            widget.setVisible(True)

        self.ToolButtons.setVisible(True)
        show = False
        count = 0
        for index in range(self.ToolButtonslayout.count()):
            if self.ToolButtonslayout.itemAt(index).widget().isVisible():
                count += 1
                if count >= 2:
                    show = True
        self.ToolButtons.setVisible(show)

        if self.canvas.tool.type not in self.canvas.painter.tools:
            self.setCanvasTool(canvasTool.drag.name)



        
    def addClass(self):
        self.classList.addClass()
        
    def removeLastClass(self):
        self.classList.removeClass()     
        
    def activeClass(self):
        c = self.classList.getActiveClass()
        assert(c >= 0)
        return c
    
    def ClassColor(self, c = None):
        if c is None:
            c = self.activeClass()
        color = self.classList.getClassColor(c)
        color.setAlpha(255)
        return color
    
    def NumOfClasses(self):
        return self.classList.getNumberOfClasses()
    
    def setNumberOfClasses(self, numclasses):
        self.classList.setNumberOfClasses(numclasses)

    def setCanvasFocus(self):
        self.canvas.setFocus()
    
    def updateImage(self):
        self.canvas.checkForChanges()
        self.setCanvasFocus()
        
    def classcolorChanged(self):
        self.canvas.redrawImage()
        self.updateCursor()

    def classChanged(self):
        self.setToolButtons()
        self.updateCursor()
        self.canvas.painter.resetSmartMask()

    def updateCursor(self):
        self.canvas.updateCursor()
        self.setCanvasFocus()

    def clear(self, resetview = True):
        self.canvas.clearLabel()
        self.deleteLabel()
        self.canvas.ReloadLabels(resetview)
        QApplication.processEvents()

    def undo(self):
        self.canvas.undoDrawing()

    def checkIfUndoPossible(self):
        if self.canvas.painter.undoPossible():
            self.Bundo.setEnabled(True)
        else:
            self.Bundo.setEnabled(False)
    
    def deleteLabel(self):
        if self.files.CurrentLabelPath() and os.path.exists(self.files.CurrentLabelPath()):
            os.remove(self.files.CurrentLabelPath())
        
    def nextImage(self):
        self.files.nextImage()
        self.changeImage()

    def previousImage(self):
        self.files.previousImage()
        self.changeImage()
            
    def changeImage(self, forcereset=False):
        if self.files.numofImages > 0:
            self.readCurrentImageAsBGR(forcereset)
        else:
            # width = self.canvas.geometry().width()
            # height = self.canvas.geometry().height()
            self.canvas.reset()
            self.currentImageFile = None
        self.setImageText() 

    def setImageText(self):
        if self.files.currentImage == -1:
            text = 'No Image Displayed'
        else:
            if not self.files.CurrentImageName() or not self.currentImageFile:
                return
            text = "%d of %i: " % (self.files.currentImage+1,self.files.numofImages) + self.files.CurrentImageName(True) 
            if self.currentImageFile.isStack():
                text = text + " " + "%d/%i: " % (self.files.currentFrame+1,self.currentImageFile.numOfImagesInStack())
        self.StatusFile.setText(text)
          
    def setTrainFolder(self):
        # folder = str(QFileDialog.getExistingDirectory(self, "Select Training Image Folder"))
        folder = openFolder("Select Training Image Folder")
        if not folder:
            return
        
        self.files.setTrainFolder(folder)
        self.setWorkingFolder()
        
    def setTestFolder(self):
        # folder = str(QFileDialog.getExistingDirectory(self, "Select Test Image Folder"))
        folder = openFolder("Select Test Image Folder")
        if not folder:
            return
        self.files.setTestFolder(folder)
        self.setWorkingFolder()
       
    def switchToTrainFolder(self):
        self.files.switchToTrainFolder()
        self.setWorkingFolder()
        
    def switchToTestFolder(self):
        self.files.switchToTestFolder()
        self.setWorkingFolder()
            
    def setWorkingFolder(self):
        self.files.getFiles()
        
        if self.files.currentimagesrootpath:
            self.loadTrack()
            model = QFileSystemModel()
            model.setRootPath(self.files.currentimagesrootpath)
            self.TVFiles.setModel(model)
            self.TVFiles.setRootIndex(model.index(self.files.currentimagesrootpath))
            self.TVFiles.setVisible(True)
        else:
            self.TVFiles.setVisible(False)
            
        self.changeImage(forcereset = True)
        self.setFolderLabels()
        
    def FileExplorerItemSelected(self, index):
        file = index.model().fileName(index)
        path = index.model().filePath(index)
        if self.files.moveToSubFolder(os.path.dirname(os.path.abspath(path))):
            self.loadTrack()
        if self.files.setImagebyName(file):
            self.changeImage()


    def ensureLabelFolder(self):
        self.files.ensureLabelFolder()


    def setFolderLabels(self):
        if self.files.trainImagespath:
            self.BTrainImageFolder.setText('...' + self.files.trainImagespath[-18:])
            self.BTrainImageFolder.setEnabled(True)
        if self.files.testImagespath:
            self.BTestImageFolder.setText('...' + self.files.testImagespath[-20:])
            self.BTestImageFolder.setEnabled(True)
        if self.files.train_test_dir:
            if self.BTrainImageFolder.isEnabled():
                self.BTrainImageFolder.setStyleSheet('font:bold; text-align:left')
                self.BTestImageFolder.setStyleSheet('font:normal;text-align:left')
        else:
            if self.BTestImageFolder.isEnabled():
                self.BTrainImageFolder.setStyleSheet('font:normal;text-align:left')
                self.BTestImageFolder.setStyleSheet('font:bold;text-align:left')
    
    def ClassListUpdated(self):
        if self.initialized:
            self.training_form.SBClasses.SpinBox.setValue(self.NumOfClasses())
            self.training_form.settings_form.changeClassWeightSettings()

    def copyStackLabels(self):
        if self.files.CurrentLabelPath() and os.path.exists(self.files.CurrentLabelPath()) and self.hasStack():
            if self.ConfirmPopup('Are you sure? All labels of the current stack are replaced.'):
                self.files.copyStackLabels(self.files.CurrentLabelPath(), self.currentImageFile.numOfImagesInStack())

    def hasStack(self):
        if self.currentImageFile is None:
            return False
        return self.currentImageFile.isStack()
            
    def hasImage(self):
        return self.currentImageFile is not None
         
    def setCanvasMode(self):
        tool = self.sender().objectName()
        self.setCanvasTool(tool)
        
    def setCanvasTool(self, tool):
        self.canvas.setnewTool(tool)

    def selectToolButtonX(self,x):
        i = 0
        for b in canvasToolButton:
            widget = self.findChild(DCDButton, b.name)
            if not widget:
                continue
            if widget.isVisible():
                i += 1
            if i == x:
                self.setCanvasTool(widget.objectName())
                return

    def readCurrentImageAsBGR(self, forcereset=False):
        with self.wait_cursor():
            
            keep = False
            if not forcereset and self.currentImageFile and self.keepImageSettings:
                br = self.currentImageFile.brightness
                c = self.currentImageFile.contrast
                keep = True
            
            try:
                self.currentImageFile = ImageFile(self.files.CurrentImagePath(), asBGR = True)
            except:
                self.PopupWarning('Can not load image')
                return
                
            self.resetBrightnessContrast()
            if self.currentImageFile.isStack():
                self.SFrame.show()
                self.initFrameSlider()
            else:
                self.SFrame.hide()
            
            if keep:
                self.currentImageFile.brightness = br
                self.currentImageFile.contrast = c
           
            self.canvas.ReloadImage(not keep)


    def initFrameSlider(self):
        maxval = self.currentImageFile.numOfImagesInStack()
        self.SFrame.setMaximum(maxval)
        self.files.currentFrame = 0
        self.SFrame.setValue(maxval)
        self.canvas.resetView()     
  
    def FrameChanged(self):
        self.files.currentFrame = self.currentImageFile.numOfImagesInStack() - self.SFrame.value() 
        self.canvas.ReloadImage(resetView = False)
        self.setImageText()
        
    def setBrightness(self):
        if not self.currentImageFile:
            return
        self.currentImageFile.brightness = self.SBrightness.value()
        self.canvas.ReloadImage(resetView = False)
        
    def setContrast(self):
        if not self.currentImageFile:
            return
        c = self.SContrast.value()
        if c > 1:
            self.currentImageFile.contrast = 1 + c/10 
        else:
            self.currentImageFile.contrast = 1 + c/51
        self.canvas.ReloadImage(resetView = False)
        
    def keepImageSettingsChanged(self):
        self.keepImageSettings = self.CBKeep.isChecked()
        
    def resetBrightnessContrast(self):
        self.SContrast.setValue(0)
        self.SBrightness.setValue(0)
        
        self.setContrast()
            
    def getCurrentImage(self):   
        if self.currentImageFile:
            return self.currentImageFile.getCorrectedImage(self.files.currentFrame)
        else:
            return None
    
    def setWorkers(self, numWorker):
        self.maxworker = numWorker
        self.dl.worker = numWorker
    
    def emitinitProgress(self, num):
        self.initProgress.emit(num)
        
    def emitProgress(self):
        self.Progress.emit()
        
    def emitProgressFinished(self):
        self.ProgressFinished.emit()
        
    def emitBlockWindow(self):
        self.Block.emit()
        
    def emitUnBlockWindow(self):
        self.UnBlock.emit()
        
    def emitPopup(self, message):
        self.Popup.emit(message)
        
    def emitStatus(self, message):
        self.StatusMessage.emit(message)
    
    
    @pyqtSlot(int)
    def initializeProgress(self, ticks):
        assert(ticks != 0)
        self.progresstick = 1/ticks *100
        self.progress = 0
        self.setProgress(0)
     
    @pyqtSlot()  
    def addProgress(self):
        with self.lock:
            self.progress += self.progresstick
            
        if self.progresslock.locked():
            return
        with self.progresslock:
            self.setProgress(self.progress)
    
    @pyqtSlot()
    def finishProgress(self):
        self.setProgress(100)
              
    @pyqtSlot()
    def blockWindow(self):
        self.setEnabled(False)
        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        QApplication.processEvents()
        
    @pyqtSlot()
    def unblockWindow(self):
        self.setEnabled(True)
        QApplication.restoreOverrideCursor()
        QApplication.processEvents()

    def setProgress(self,value):
        if value < 100:
            self.StatusProgress.show()
            self.StatusProgress.setValue(value)
        else:
            self.StatusProgress.hide()
        self.processEvents()

    def processEvents(self):
        if self.processlock.locked():
            return
        with self.processlock:
            QApplication.processEvents()
    
    def writeStatus(self,msg):
        self.Status.setText(msg)

            
    def toggleTrainStatus(self, training):
        self.lockModel(training)
        self.training_form.setParameterStatus()
        self.training_form.toggleTrainStatus(training=training)
        self.plotting_form.toggleTrainStatus(training=training)

    def togglePredictionStatus(self, prediction):
        self.lockModel(prediction)
        self.training_form.toggleTrainStatus(training=prediction)

    def lockModel(self, lock):
        self.Bloadmodel.setEnabled(not lock)
        self.Bresetmodel.setEnabled(not lock)
    
    def setClassWeights(self): 
        if self.NumOfClasses() > 2:
            weights = np.zeros((self.NumOfClasses()), dtype = np.float)
            for i in range(self.NumOfClasses()):
                weights[i] = self.classList.getClassWeight(i)
        else:
            weights = self.classList.getClassWeight(1)            
        self.dl.data.setClassWeights(weights)
        

    ## deep learning  
    def LearningMode(self):
        return self.dl.WorkingMode

    def loadModel(self):
        if self.saveload_modelweights_only:
            filename = loadFile("Select Model File", "Model (*.h5)")
            if filename:
                with self.wait_cursor():
                    try:
                        self.settings.loadSettings(filename.replace("h5", "json"))
                    except:
                        pass
                    self.dl.initModel(self.NumOfClasses(), self.training_form.CBMono.isChecked())
                    self.dl.LoadModelWeights(filename)
                    self.writeStatus('model loaded')
                    
        else:
            filename = openFolder('Select a Model Directory')
            if filename:
                with self.wait_cursor():
                    self.dl.LoadModel(filename)
                    self.writeStatus('model loaded')
                    try:
                        settings_file = os.path.join(filename, 'settings.json')
                        self.settings.loadSettings(settings_file)
                    except:
                        pass


    def saveModel(self):
        if not self.dl.initialized:
            self.PopupWarning('No model trained/loaded')
            return
            

        if self.saveload_modelweights_only:
            filename = saveFile("Save Model To File","Model File (*.h5)", 'h5')
            if filename:
                with self.wait_cursor():
                    self.dl.SaveModelWeights(filename)
                    self.settings.saveSettings(filename.replace("h5", "json"), networksettingsonly=True)
                    self.writeStatus('model saved')
        else: 
            filename = openFolder('Select a Folder as Model Directory')
            
            if filename:
                with self.wait_cursor():
                    self.dl.SaveModel(filename)
                    settings_file = os.path.join(filename, 'settings.json')
                    self.settings.saveSettings(settings_file, networksettingsonly=True)
                    self.writeStatus('model saved')

    def resetModel(self):
        self.dl.Reset()
        self.training_form.setParameterStatus()
        self.writeStatus('model reset')
    
    def showTrainingWindow(self):
        if self.files.trainImagespath is None:
            self.PopupWarning('Training folder not selected')
            return
 
        self.training_form.show()
        
    def startTraining(self):
        if not self.dl.parameterFit(self.NumOfClasses(), self.training_form.CBMono.isChecked()):
            if not self.dl.initModel(self.NumOfClasses(), self.training_form.CBMono.isChecked()):
               self.PopupWarning('Cannot initialize model') 
               return
        self.plotting_form.show()
        self.plotting_form.initialize()
        self.toggleTrainStatus(True)
        self.dl.startTraining(self.files.trainImagespath, self.files.validationpath)

    def resumeTraining(self):
        self.toggleTrainStatus(True)
        if not self.dl.resumeTraining():
            self.toggleTrainStatus(False)
        
    def predictAllImages(self):
        if self.files.testImagespath is None:
            self.PopupWarning('Prediction folder not selected')
            return

        if not self.dl.initialized:
            self.PopupWarning('No model trained/loaded')
            return
        if not self.ConfirmPopup('Are you sure to predict all images in the test folder?'):
            return

        self.emitinitProgress(len(self.files.TestImages))
        self.togglePredictionStatus(True)
        thread = WorkerThread(self.predictAllImages_async, callback_end=self.PredictionFinished)
        thread.daemon = True
        thread.start()


    def PredictionFinished(self):
        self.writeStatus(str(len(self.files.TestImages)) + ' images predicted')
        self.canvas.ReloadImage()
        self.togglePredictionStatus(False)

    def predictAllImages_async(self):
        for image in self.files.TestImages:
            self.predictSingleImageOrStack(image)
        self.emitProgressFinished()

    def predictAllImages_concurrent(self):
        self.emitinitProgress(len(self.files.TestImages))
        with concurrent.futures.ThreadPoolExecutor(max_workers=self.maxworker) as executor:
            executor.map(self.predictSingleImageOrStack,images)

        self.emitProgressFinished()
        self.switchToTestFolder()
        self.writeStatus(str(len(images)) + ' images predicted')
        
    def predictSingleImageOrStack(self, imagepath):
        # in worker thread
        image = ImageFile(imagepath)
        image.normalizeImage()
        name = self.files.getFilenameFromPath(imagepath)
        if image.isStack():
            path = self.files.extendLabelPathByFolder(self.files.testImageLabelspath, name)
        else:
            path = self.files.testImageLabelspath
        for i in range(image.numOfImagesInStack()):
            img = image.getDLInputImage(self.dl.MonoChrome, i)
            pred = self.dl.Mode.PredictImage(img)
            if image.isStack():
                filename = self.files.extendNameByFrameNumber(name,i)
            else:
                filename = name
            self.extractAndSavePredictedResults(pred, path, filename)
        self.emitProgress()
        
    def predictImage_async(self):
        image = self.dl.data.readImage(self.files.CurrentImagePath(), self.files.currentFrame)
        if image is None:
            return
        self.clear(resetview = False)
        image = self.canvas.getFieldOfViewImage(image)
        self.dl.PredictImage_async(image)

        
    def predictImage(self):
        if self.files.CurrentImagePath() is None:
            return
        if not self.dl.initialized:
            self.PopupWarning('No model trained/loaded')
            return
        with self.wait_cursor():
            image = self.dl.data.readImage(self.files.CurrentImagePath(), self.files.currentFrame)
            image = self.canvas.getFieldOfViewImage(image)
            
            if image is None:
                self.PopupWarning('Cannot load image')
                return
            prediction = self.dl.Mode.PredictImage(image)
            if prediction is None:
                self.PopupWarning('Cannot predict image')
                return

            self.extractAndUpdatePredictedResults(prediction)
            self.writeStatus('image predicted')
            
    def extractPredictedResults(self, prediction, offset = (0,0)):
        if prediction is None:
            return
        return self.dl.Mode.extractShapesFromPrediction(prediction, not self.allowInnerContours, offset)
            
    def extractAndSavePredictedResults(self, prediction, path, name):
        if prediction is None:
            return
        self.files.ensureFolder(path)
        self.dl.Mode.saveShapes(self.extractPredictedResults(prediction), os.path.join(path, (name + ".npz")))
        
    def extractAndUpdatePredictedResults(self, prediction):
        self.canvas.painter.clearFOV()
        fov = self.canvas.getFieldOfViewRect()
        shapes = self.extractPredictedResults(prediction, offset = (int(fov.x()),int(fov.y())))
        self.updateCurrentShapes(shapes)

        
    def updateCurrentShapes(self, newshapes):
        if newshapes:
            self.canvas.painter.addShapes(newshapes)
            self.canvas.checkForChanges()

    def autoSegment(self):
        with self.wait_cursor():
            self.canvas.painter.autosegment()
    
    def autoPointDetection(self):
        with self.wait_cursor():
            self.canvas.painter.autodetection()
        
    def saveDLClass(self):
        
        defaultname = 'model_' + time.strftime("%Y-%m-%d_%H-%M-%S")
        filename = saveFile("Save Settings","Settings File (*.pkl)", 'pkl', defaultname)
        if filename:
            filehandler = open(filename, 'wb') 
            self.settings.saveSettings(filename.replace("pkl", "json"), networksettingsonly=True)
            
            self.dl.detachObserver(self.dlobserver)
            hed = self.dl.hed
            self.dl.hed = None
            model = self.dl.Model
            self.dl.Model = None
            
            pickle.dump(self.dl, filehandler)
            self.dl.hed = hed
            self.dl.Model = model
            self.dl.attachObserver(self.dlobserver)
        
    # deep learning observer functions
    def dlStarted(self):
        self.toggleTrainStatus(True)
        
    def dlProgress(self):
        self.plotting_form.refresh()
        
    def dlEpoch(self, epoch):
        if self.files.train_test_dir or self.predictionRate == 0:
            return    
        if (epoch % self.predictionRate) == self.predictionRate - 1:
            self.predictImage_async()
        
    def dlFinished(self):
        self.toggleTrainStatus(False)

    def dlResult(self, result):
        self.canvas.painter.clearFOV()
        self.extractAndUpdatePredictedResults(result)
        # self.clear()
        # self.extractAndSavePredictedResults(result, self.files.labelpath, self.files.CurrentLabelName())
        # self.canvas.ReloadImage()

    def dlError(self, err):
        (exctype, value, traceback) = err
        if exctype == errors.ResourceExhaustedError:
            self.PopupWarning('Out of memory!') 
        else:
            MIA_UI.excepthook(exctype, value, traceback)

              
    ############################
    @contextmanager
    def wait_cursor(self):
        try:
            QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
            QApplication.processEvents()
            yield
        finally:
            QApplication.restoreOverrideCursor()

    def excepthook(excType, excValue, tracebackobj):
        
        timeString = time.strftime("%Y-%m-%d, %H:%M:%S")
        separator = '-' * 80
        
        tb = io.StringIO()
        traceback.print_tb(tracebackobj, None, tb)
        tb.seek(0)
        info = tb.read()
        errmsg = '%s: \n%s' % (str(excType), str(excValue))
        sections = [separator, timeString, separator, errmsg, separator, info]
        msg = '\n'.join(sections)
        print(msg)
        
        try:
            f = open(LOG_FILENAME, "a")
            f.write(msg)
            f.close()
        except IOError:
            pass
        
        msg = QMessageBox()
        msg.setIcon(QMessageBox.Critical)
        msg.setText('An error occured, a logfile was created with further information. \nPlease forward logfile. \n')
        msg.setWindowTitle("Error")
        msg.setStandardButtons(QMessageBox.Ok)
        msg.exec_()
    
sys.excepthook = MIA_UI.excepthook



# -*- coding: utf-8 -*-
"""
Created on Wed Jan 29 13:33:38 2020

@author: Koerber
"""

from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
from PyQt5.QtCore import *
from ui.ui_Augment import AugmentWindow
from ui.ui_TrainSettings import TrainSettingsWindow
from ui.segmentation.ui_SegSettings import SegmentationSettingsWindow
from ui.objectdetection.ui_ODSettings import ObjectDetectionSettingsWindow
from ui.ui_utils import LabelledAdaptiveDoubleSpinBox, LabelledSpinBox, LabelledDoubleSpinBox, DCDButton
from ui.style import styleForm
from dl.method.mode import dlMode


class Window(object):
    def setupUi(self, Form):
        Form.setWindowTitle('Neural Network Training') 

        styleForm(Form)
        self.centralWidget = QWidget(Form)
    
        self.vlayout = QVBoxLayout(self.centralWidget)
        self.vlayout.setContentsMargins(3, 10, 3, 3)
        self.vlayout.setSpacing(2)
        self.centralWidget.setLayout(self.vlayout)
        
        # row 0
        self.Model = self.ModelGroup()
        self.vlayout.addWidget(self.Model)
        
        self.Parameter = self.ParameterGroup()
        self.vlayout.addWidget(self.Parameter)

        # row 4
        self.settings3layout = QHBoxLayout(self.centralWidget)
        self.BModeSettings = DCDButton(self.centralWidget, 'Extended')
        self.BModeSettings.setToolTip('Open extended settings')
        self.BModeSettings.setIcon(QIcon('icons/settings.png'))
        self.settings3layout.addWidget(self.BModeSettings)
        self.BSettings = DCDButton(self.centralWidget, 'Settings')
        self.BSettings.setToolTip('Open training settings')
        self.BSettings.setIcon(QIcon('icons/settings.png'))
        self.settings3layout.addWidget(self.BSettings)
        self.BAugmentation = DCDButton(self.centralWidget, 'Augmentations')
        self.BAugmentation.setToolTip('Open image augmentation settings')
        self.BAugmentation.setIcon(QIcon('icons/augmentation.png'))
        self.settings3layout.addWidget(self.BAugmentation)
        self.vlayout.addLayout(self.settings3layout)
        
        # row 5
        self.BTrain = DCDButton(self.centralWidget, 'Start Training')
        self.BTrain.setIcon(QIcon('icons/train.png'))
        self.BTrain.setToolTip('Start training with selected parameters')
        self.vlayout.addWidget(self.BTrain)  


        self.centralWidget.setFixedSize(self.vlayout.sizeHint())
        Form.setFixedSize(self.vlayout.sizeHint())


    def ModelGroup(self):
        groupBox = QGroupBox("Deep Learning Model")
        layout = QVBoxLayout(self.centralWidget)

        hlayout = QHBoxLayout(self.centralWidget)
        self.CBArchitecture = QComboBox(self.centralWidget)
        self.CBArchitecture.setObjectName('nn_Architecture')
        self.CBArchitecture.setToolTip('Select model architecture for training')
        self.LArchitecture = QLabel(self.centralWidget)
        self.LArchitecture.setText('Deep Learning Architecture')
        hlayout.addWidget(self.CBArchitecture)
        hlayout.addWidget(self.LArchitecture)
     
        layout.addLayout(hlayout)

        h2layout = QHBoxLayout(self.centralWidget)
        self.CBBackbone = QComboBox(self.centralWidget)
        self.CBBackbone.setObjectName('nn_Backbone')
        self.CBBackbone.setToolTip('Select model Backbone')
        self.LBackbone = QLabel(self.centralWidget)
        self.LBackbone.setText('Model Backbone')
        h2layout.addWidget(self.CBBackbone)
        h2layout.addWidget(self.LBackbone)
     
        layout.addLayout(h2layout)

        self.CBPretrained = QCheckBox("Pretrained Net",self.centralWidget)
        self.CBPretrained.setObjectName('nn_preTrained')
        self.CBPretrained.setToolTip('Select to use an imagenet-pretrained model')
        layout.addWidget(self.CBPretrained)
        groupBox.setLayout(layout)
        return groupBox


    def ParameterGroup(self):
        groupBox = QGroupBox("Training Parameter")
        layout = QVBoxLayout(self.centralWidget)
        # row 1
        self.settings0layout = QHBoxLayout(self.centralWidget)
        self.SBClasses = LabelledSpinBox('Classes', self.centralWidget)
        self.SBClasses.setObjectName('nn_NumOfClasses')
        self.SBClasses.SpinBox.setRange(1,100)
        self.SBClasses.setToolTip('Number of classes')
        self.SBClasses.SpinBox.setEnabled(False)

        self.CBMono = QRadioButton(self.centralWidget)
        self.CBMono.setText("Mono")
        self.CBMono.setObjectName('nn_MonochromaticInput')
        self.CBMono.setToolTip('Select if only monochromatic images are used as input')
        
        self.CBRGB = QRadioButton(self.centralWidget)
        self.CBRGB.setText("Color")
        self.CBRGB.setObjectName('nn_RGBInput')
        self.CBRGB.setToolTip('Select if color images are used as input')
        self.settings0layout.addWidget(self.SBClasses,stretch = 2)
        self.settings0layout.addWidget(self.CBRGB, stretch = 1)
        self.settings0layout.addWidget(self.CBMono, stretch = 1)
        layout.addLayout(self.settings0layout)
        
        # row 2
        self.settings1layout = QHBoxLayout(self.centralWidget)
        self.SBBatchSize = LabelledSpinBox('Batch Size',self.centralWidget)
        self.SBBatchSize.setToolTip('Set batch size')
        self.SBBatchSize.setObjectName('nn_BatchSize')
        self.SBBatchSize.SpinBox.setRange(1,2048)

        self.SBEpochs = LabelledSpinBox('Epochs',self.centralWidget)
        self.SBEpochs.setToolTip('Set number of epochs')
        self.SBEpochs.setObjectName('nn_Epochs')
        self.SBEpochs.SpinBox.setRange(1,10000)
        self.settings1layout.addWidget(self.SBBatchSize)
        self.settings1layout.addWidget(self.SBEpochs)

        layout.addLayout(self.settings1layout)
        
        # row 3
        self.settings2layout = QHBoxLayout(self.centralWidget)
        self.SBLearningRate = LabelledAdaptiveDoubleSpinBox ('Learning Rate',self.centralWidget)
        self.SBLearningRate.SpinBox.setRange(0.000001,1)
        self.SBLearningRate.setToolTip('Set learning rate')
        self.SBLearningRate.SpinBox.setDecimals(6)
        self.SBLearningRate.SpinBox.setSingleStep(0.1)

        self.SBScaleFactor = LabelledDoubleSpinBox('Down Scaling',self.centralWidget)
        self.SBScaleFactor.SpinBox.setRange(0.1,1)
        self.SBScaleFactor.setObjectName('nn_ImageDownScaling')
        self.SBScaleFactor.setToolTip('Set image reduction factor')
        self.SBScaleFactor.SpinBox.setSingleStep(0.1)
        self.SBScaleFactor.SpinBox.setDecimals(1)

        self.settings2layout.addWidget(self.SBLearningRate)
        self.settings2layout.addWidget(self.SBScaleFactor)
        layout.addLayout(self.settings2layout)
        groupBox.setLayout(layout)
        return groupBox
        

class TrainingWindow(QMainWindow, Window):
    def __init__(self, parent ):
        super(TrainingWindow, self).__init__(parent)
        self.parent = parent
        self.setupUi(self)   
        
        self.SBLearningRate.SpinBox.setMinimum(self.parent.dl.lrschedule.minlr)
        self.SBLearningRate.SpinBox.setValue(self.parent.dl.learning_rate)
        self.CBMono.setChecked(True)
        self.SBEpochs.SpinBox.setValue(self.parent.dl.epochs)
        self.SBScaleFactor.SpinBox.setValue(self.parent.dl.ImageScaleFactor)
        self.SBBatchSize.SpinBox.setValue(self.parent.dl.batch_size)
        
        
        
        self.SBBatchSize.SpinBox.valueChanged.connect(self.BatchSizeChanged)
        self.SBEpochs.SpinBox.valueChanged.connect(self.EpochsChanged)
        self.SBLearningRate.SpinBox.valueChanged.connect(self.LRChanged)
        self.SBScaleFactor.SpinBox.valueChanged.connect(self.ScaleFactorChanged)
        self.CBArchitecture.currentIndexChanged.connect(self.ArchitectureChanged)
        self.CBBackbone.currentIndexChanged.connect(self.BackBoneChanged)
        self.CBPretrained.setChecked(self.parent.dl.Mode.pretrained)
        self.CBPretrained.stateChanged.connect(self.UsePretrained)

        self.setModelOptions()

        self.settings_form = TrainSettingsWindow(self)
        self.augmentation_form = AugmentWindow(self.parent)
        # training mode settings
        self.segmentationssettings_form = SegmentationSettingsWindow(self)
        self.objectdetectionsettings_form = ObjectDetectionSettingsWindow(self)
        

        self.BSettings.clicked.connect(self.settings_form.show)
        self.BAugmentation.clicked.connect(self.augmentation_form.show)
        self.BModeSettings.clicked.connect(self.showModeSettings)
        self.BTrain.clicked.connect(self.Train)
        
    def show(self):
        self.setParameterStatus()
        self.ModeChanged()
        super(TrainingWindow, self).show()
        
    def setParameterStatus(self):
        if not self.parent.dl.initialized:
            self.CBMono.setEnabled(True)
            self.CBRGB.setEnabled(True)
            self.CBArchitecture.setEnabled(True)
            self.CBBackbone.setEnabled(True)
            self.CBPretrained.setEnabled(True)
        else:
            self.CBMono.setEnabled(False)
            self.CBRGB.setEnabled(False)
            self.CBArchitecture.setEnabled(False)
            self.CBBackbone.setEnabled(False)
            self.CBPretrained.setEnabled(False)

            if self.parent.dl.MonoChrome:
                self.CBMono.setChecked(True)
            else:
                self.CBRGB.setChecked(True) 
        
    def hide(self):
        self.closeWindows()
        super(TrainingWindow, self).hide()
        
    def closeEvent(self, event):       
        self.closeWindows()
        super(TrainingWindow, self).closeEvent(event)
        
    def closeWindows(self):
        self.settings_form.hide()
        self.augmentation_form.hide()
        self.segmentationssettings_form.hide()
        self.objectdetectionsettings_form.hide()
        
    def showModeSettings(self):
        if self.parent.LearningMode() == dlMode.Segmentation:
            self.segmentationssettings_form.show()
        elif self.parent.LearningMode() == dlMode.Object_Counting:
            self.objectdetectionsettings_form.show()    
    
    def ModeChanged(self):
        self.settings_form.updateLossesAndMetrics()
        self.closeWindows()
        self.setModelOptions()
        self.UsePretrained()
        self.BModeSettings.setEnabled(True)
        if self.parent.LearningMode() == dlMode.Segmentation:
            self.BModeSettings.setToolTip('Open segmentation settings')
            self.BModeSettings.setText('Segmentation')
        elif self.parent.LearningMode() == dlMode.Object_Counting:
            self.BModeSettings.setToolTip('Open object detection settings')
            self.BModeSettings.setText('Detection')
        elif self.parent.LearningMode() == dlMode.Classification:
            self.BModeSettings.setToolTip('Open classification settings')
            self.BModeSettings.setText('Classification')
            self.BModeSettings.setEnabled(False)


    def setModelOptions(self):
        self.setArchitectureOptions()
        self.setBackBoneOptions()
        
    def setArchitectureOptions(self):
        defaultarchitecture= self.parent.dl.Mode.architecture
        defaultbackbone = self.parent.dl.Mode.backbone
        self.CBArchitecture.clear()
        for arch in self.parent.dl.Mode.getArchitectures():
            self.CBArchitecture.addItem(arch)   
        self.CBArchitecture.setCurrentIndex(self.CBArchitecture.findText(defaultarchitecture))
        self.CBBackbone.setCurrentIndex(self.CBBackbone.findText(defaultbackbone))

    def setBackBoneOptions(self):
        defaultbackbone = self.parent.dl.Mode.backbone
        self.CBBackbone.clear()
        for bb in self.parent.dl.Mode.getBackbones():
            self.CBBackbone.addItem(bb) 
        self.CBBackbone.setCurrentIndex(self.CBBackbone.findText(defaultbackbone))

    def ArchitectureChanged(self):
        if self.CBArchitecture.currentText() != '':
            self.parent.dl.Mode.setArchitecture(self.CBArchitecture.currentText())
        self.setBackBoneOptions()

    def BackBoneChanged(self):
        if self.CBBackbone.currentText() != '':
            self.parent.dl.Mode.setBackbone(self.CBBackbone.currentText())

    def ModelType(self):
        if self.CBArchitecture.currentText() != '':
            self.parent.dl.Mode.setArchitecture(self.CBArchitecture.currentText())
        if self.CBBackbone.currentText() != '':
            self.parent.dl.Mode.setBackbone(self.CBBackbone.currentText())

    def UsePretrained(self):
        self.parent.dl.Mode.pretrained = self.CBPretrained.isChecked()
        
    def BatchSizeChanged(self):
        self.parent.dl.batch_size = self.SBBatchSize.SpinBox.value()
        
    def EpochsChanged(self):
        self.parent.dl.epochs = self.SBEpochs.SpinBox.value()
        
    def LRChanged(self):
        self.parent.dl.learning_rate = self.SBLearningRate.SpinBox.value()
        self.settings_form.silentlyUpdateLR(self.SBLearningRate.SpinBox.value())      
        
    def ScaleFactorChanged(self):
        self.parent.dl.ImageScaleFactor = self.SBScaleFactor.SpinBox.value()
        self.parent.settings_form.silentlyUpdateScale(self.SBScaleFactor.SpinBox.value())
  
    def Train(self):
        self.parent.startTraining()
        
    def toggleTrainStatus(self, training):
        self.BTrain.setEnabled(not training)
        self.setEnabled(not training)
        self.settings_form.setEnabled(not training)
        self.augmentation_form.setEnabled(not training)


        
        


# -*- coding: utf-8 -*-
"""
Created on Wed Aug 12 14:30:53 2020

@author: Koerber
"""


from abc import ABC, abstractmethod
from enum import Enum

class dlLoss(Enum):
    focal_loss = 1
    cross_entropy = 2
    dice_loss = 3
    kl_divergence = 4
    mean_squared_err = 5
    mean_abs_err = 6
    mean_squared_log_err = 7


class Losses(ABC):
    def __init__(self, parent):
        self.supportedlosses = []
        self.parent = parent
        self._loss = None
        
    @abstractmethod
    def getLoss(self):
        pass
    
    @property
    def loss(self):
        return self._loss
    
    @loss.setter
    def loss(self, loss):
        self._loss = loss
        
    
    
